# pylint: disable=too-many-lines
import argparse
import calendar
import datetime
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

import annofabapi
import more_itertools
import numpy
import pandas
from annofabapi.models import OrganizationMember, Project, ProjectMember
from annofabapi.utils import allow_404_error
from dataclasses_json import DataClassJsonMixin
from more_itertools import first_true

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login, get_list_from_args
from annofabcli.common.utils import isoduration_to_hour

logger = logging.getLogger(__name__)


def _create_required_columns(df: pandas.DataFrame, prior_columns: List[Any]) -> List[str]:
    remained_columns = list(df.columns.difference(prior_columns))
    all_columns = prior_columns + remained_columns
    return all_columns


def catch_exception(function: Callable[..., Any]) -> Callable[..., Any]:
    """
    Exceptionをキャッチしてログにstacktraceを出力する。
    """

    def wrapped(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(e)
            logger.exception(e)

    return wrapped


@dataclass(frozen=True)
class User(DataClassJsonMixin):
    account_id: str
    user_id: str
    username: str
    biography: Optional[str]


@dataclass(frozen=True)
class LaborWorktime(DataClassJsonMixin):
    """
    労務管理情報
    """

    date: str
    organization_id: str
    organization_name: str
    project_id: str
    project_title: str
    account_id: str
    user_id: str
    username: str
    biography: Optional[str]
    worktime_plan_hour: float
    """労務管理画面の予定作業時間"""
    worktime_result_hour: float
    """労務管理画面の実績作業時間"""
    worktime_monitored_hour: Optional[float]
    """AnnoFabの作業時間"""
    working_description: Optional[str]
    """実績作業時間に対する備考"""


@dataclass(frozen=True)
class LaborAvailability(DataClassJsonMixin):
    """
    労務管理情報
    """

    date: str
    account_id: str
    user_id: str
    username: str
    availability_hour: float


@dataclass(frozen=True)
class SumLaborWorktime(DataClassJsonMixin):
    """
    出力用の作業時間情報
    """

    date: str
    user_id: str
    worktime_plan_hour: float
    worktime_result_hour: float


class ListWorktimeByUserMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    DATE_FORMAT = "%Y-%m-%d"
    MONTH_FORMAT = "%Y-%m"

    _dict_account_statistics: Dict[str, List[Dict[str, Any]]] = {}
    """project_idごとの統計情報dict"""

    @allow_404_error
    def _get_account_statistics(self, project_id) -> Optional[List[Any]]:
        account_statistics = self.service.wrapper.get_account_statistics(project_id)
        return account_statistics

    def _get_worktime_monitored_hour_from_project_id(
        self, project_id: str, account_id: str, date: str
    ) -> Optional[float]:
        account_statistics = self._dict_account_statistics.get(project_id)
        if account_statistics is None:
            result = self._get_account_statistics(project_id)
            if result is not None:
                account_statistics = result
            else:
                logger.warning(f"project_id={project_id}: プロジェクトにアクセスできないため、アカウント統計情報を取得できませんでした。")
                account_statistics = []

            self._dict_account_statistics[project_id] = account_statistics

        return self._get_worktime_monitored_hour(account_statistics, account_id=account_id, date=date)

    @staticmethod
    def _get_worktime_monitored_hour(
        account_statistics: List[Dict[str, Any]], account_id: str, date: str
    ) -> Optional[float]:
        """
        AnnoFabの作業時間を取得する。
        """
        stat = first_true(account_statistics, pred=lambda e: e["account_id"] == account_id)
        if stat is None:
            return None
        histories = stat["histories"]
        hist = first_true(histories, pred=lambda e: e["date"] == date)
        if hist is None:
            return None
        return isoduration_to_hour(hist["worktime"])

    @staticmethod
    def create_required_columns(df, prior_columns):
        remained_columns = list(df.columns.difference(prior_columns))
        all_columns = prior_columns + remained_columns
        return all_columns

    @staticmethod
    def get_member_from_user_id(
        organization_member_list: List[OrganizationMember], user_id: str
    ) -> Optional[OrganizationMember]:
        member = more_itertools.first_true(organization_member_list, pred=lambda e: e["user_id"] == user_id)
        return member

    @staticmethod
    def get_project_title(project_list: List[Project], project_id: str) -> str:
        project = more_itertools.first_true(project_list, pred=lambda e: e["project_id"] == project_id)
        if project is not None:
            return project["title"]
        else:
            return ""

    @staticmethod
    def get_worktime_hour(working_time_by_user: Optional[Dict[str, Any]], key: str) -> float:
        if working_time_by_user is None:
            return 0

        value = working_time_by_user.get(key)
        if value is None:
            return 0
        else:
            return value / 3600 / 1000

    @staticmethod
    def _get_working_description(working_time_by_user: Optional[Dict[str, Any]]) -> Optional[str]:
        if working_time_by_user is None:
            return None

        return working_time_by_user.get("description")

    @staticmethod
    def _get_labor_worktime(
        labor: Dict[str, Any],
        member: Optional[ProjectMember],
        project_title: str,
        organization_name: str,
        worktime_monitored_hour: Optional[float],
    ) -> LaborWorktime:
        new_labor = LaborWorktime(
            date=labor["date"],
            organization_id=labor["organization_id"],
            organization_name=organization_name,
            project_id=labor["project_id"],
            project_title=project_title,
            account_id=labor["account_id"],
            user_id=member["user_id"] if member is not None else labor["account_id"],
            username=member["username"] if member is not None else labor["account_id"],
            biography=member["biography"] if member is not None else None,
            worktime_plan_hour=labor["plan_worktime"] if labor["plan_worktime"] is not None else 0,
            worktime_result_hour=labor["actual_worktime"] if labor["actual_worktime"] is not None else 0,
            working_description=labor["working_description"],
            worktime_monitored_hour=worktime_monitored_hour,
        )
        return new_labor

    def get_labor_availability_list_dict(
        self,
        user_list: List[User],
        start_date: str,
        end_date: str,
    ) -> Dict[str, List[LaborAvailability]]:
        """
        予定稼働時間を取得する
        Args:
            user_list:
            start_date:
            end_date:

        Returns:

        """
        labor_availability_dict = {}
        for user in user_list:
            # 予定稼働時間を取得するには、特殊な組織IDを渡す
            labor_list = self.service.wrapper.get_labor_control_availability(
                from_date=start_date,
                to_date=end_date,
                account_id=user.account_id,
            )
            new_labor_list = []
            for labor in labor_list:
                new_labor = LaborAvailability(
                    date=labor["date"],
                    account_id=labor["account_id"],
                    user_id=user.user_id,
                    username=user.username,
                    availability_hour=labor["availability"] if labor["availability"] is not None else 0,
                )
                new_labor_list.append(new_labor)
            labor_availability_dict[user.user_id] = new_labor_list
        return labor_availability_dict

    def get_labor_list_from_project_id(
        self,
        project_id: str,
        account_id_list: Optional[List[str]],
        start_date: Optional[str],
        end_date: Optional[str],
        add_monitored_worktime: bool = False,
    ) -> List[LaborWorktime]:
        organization, _ = self.service.api.get_organization_of_project(project_id)
        organization_name = organization["organization_name"]

        if account_id_list is None:
            logger.debug(f"project_id={project_id}の、すべての労務管理情報を取得しています。")
            labor_list = self.service.wrapper.get_labor_control_worktime(
                project_id=project_id,
                organization_id=organization["organization_id"],
                from_date=start_date,
                to_date=end_date,
            )
        else:
            labor_list = []
            for account_id in account_id_list:
                logger.debug(f"project_id={project_id}の、ユーザ{len(account_id_list)}件分の労務管理情報を取得しています。")
                tmp_labor_list = self.service.wrapper.get_labor_control_worktime(
                    project_id=project_id, from_date=start_date, to_date=end_date, account_id=account_id
                )
                labor_list.extend(tmp_labor_list)

        project_title = self.service.api.get_project(project_id)[0]["title"]
        labor_list = [
            e
            for e in labor_list
            if (e["actual_worktime"] is not None and e["actual_worktime"] > 0)
            or (e["plan_worktime"] is not None and e["plan_worktime"] > 0)
        ]
        logger.info(f"'{project_title}'プロジェクト('{project_id}')の労務管理情報の件数: {len(labor_list)}")

        new_labor_list = []
        for labor in labor_list:
            # 個人に紐付かないデータの場合
            if labor["account_id"] is None:
                continue

            member = self.facade.get_project_member_from_account_id(labor["project_id"], labor["account_id"])
            if add_monitored_worktime:
                try:
                    worktime_monitored_hour = self._get_worktime_monitored_hour_from_project_id(
                        project_id=project_id, account_id=labor["account_id"], date=labor["date"]
                    )
                except Exception:  # pylint: disable=broad-except
                    logger.warning(f"project_id={project_id}: 計測作業時間を取得できませんでした。", exc_info=True)
                    worktime_monitored_hour = None
            else:
                worktime_monitored_hour = None

            new_labor = self._get_labor_worktime(
                labor,
                member=member,
                project_title=project_title,
                organization_name=organization_name,
                worktime_monitored_hour=worktime_monitored_hour,
            )
            new_labor_list.append(new_labor)

        return new_labor_list

    def get_labor_list_from_organization_name(
        self,
        organization_name: str,
        account_id_list: Optional[List[str]],
        start_date: Optional[str],
        end_date: Optional[str],
        add_monitored_worktime: bool = False,
    ) -> List[LaborWorktime]:
        organization, _ = self.service.api.get_organization(organization_name)
        organization_id = organization["organization_id"]

        if account_id_list is None:
            logger.debug(f"organization_name={organization_name}の、すべての労務管理情報を取得しています。")
            labor_list = self.service.wrapper.get_labor_control_worktime(
                organization_id=organization_id, from_date=start_date, to_date=end_date
            )
        else:
            labor_list = []
            logger.debug(f"organization_name={organization_name}の、ユーザ{len(account_id_list)}件分の労務管理情報を取得しています。")
            for account_id in account_id_list:
                tmp_labor_list = self.service.wrapper.get_labor_control_worktime(
                    organization_id=organization_id, from_date=start_date, to_date=end_date, account_id=account_id
                )
                labor_list.extend(tmp_labor_list)

        labor_list = [
            e
            for e in labor_list
            if (e["actual_worktime"] is not None and e["actual_worktime"] > 0)
            or (e["plan_worktime"] is not None and e["plan_worktime"] > 0)
        ]

        logger.info(f"'{organization_name}'組織の労務管理情報の件数: {len(labor_list)}")
        project_list = self.service.wrapper.get_all_projects_of_organization(organization_name)

        new_labor_list = []
        for labor in labor_list:
            try:
                member = self.facade.get_project_member_from_account_id(labor["project_id"], labor["account_id"])
            except Exception:  # pylint: disable=broad-except
                logger.warning(f"project_id={labor['project_id']}: メンバ一覧を取得できませんでした。", exc_info=True)
                member = None

            project_title = self.get_project_title(project_list, labor["project_id"])
            if add_monitored_worktime:
                try:
                    worktime_monitored_hour = self._get_worktime_monitored_hour_from_project_id(
                        project_id=labor["project_id"], account_id=labor["account_id"], date=labor["date"]
                    )
                except Exception:  # pylint: disable=broad-except
                    logger.warning(f"project_id={labor['project_id']}: 計測作業時間を取得できませんでした。", exc_info=True)
                    worktime_monitored_hour = None
            else:
                worktime_monitored_hour = None

            new_labor = self._get_labor_worktime(
                labor,
                member=member,
                project_title=project_title,
                organization_name=organization_name,
                worktime_monitored_hour=worktime_monitored_hour,
            )
            new_labor_list.append(new_labor)

        return new_labor_list

    @staticmethod
    def get_sum_worktime_list(
        labor_list: List[LaborWorktime], user_id: str, start_date: str, end_date: str
    ) -> List[SumLaborWorktime]:
        sum_labor_list = []
        for date in pandas.date_range(start=start_date, end=end_date):
            str_date = date.strftime(ListWorktimeByUserMain.DATE_FORMAT)
            filtered_list = [e for e in labor_list if e.user_id == user_id and e.date == str_date]
            worktime_plan_hour = sum([e.worktime_plan_hour for e in filtered_list])
            worktime_result_hour = sum([e.worktime_result_hour for e in filtered_list])

            labor = SumLaborWorktime(
                user_id=user_id,
                date=date,
                worktime_plan_hour=worktime_plan_hour,
                worktime_result_hour=worktime_result_hour,
            )
            sum_labor_list.append(labor)

        return sum_labor_list

    @staticmethod
    @catch_exception
    def write_sum_worktime_list(sum_worktime_df: pandas.DataFrame, output_dir: Path):
        sum_worktime_df.round(3).to_csv(str(output_dir / "ユーザごとの作業時間.csv"), encoding="utf_8_sig", index=False)

    @staticmethod
    @catch_exception
    def write_sum_plan_worktime_list(sum_worktime_df: pandas.DataFrame, output_dir: Path) -> None:
        """
        出勤予定かどうかを判断するため、作業予定時間が"0"のときは"☓",　そうでないときは"○"で出力する
        Args:
            sum_worktime_df:
            output_dir:

        """

        def create_mark(value) -> str:
            if value == 0:
                return "×"
            else:
                return "○"

        def is_plan_column(c) -> bool:
            c1, c2 = c
            if c1 in ["date", "dayofweek"]:
                return False
            return c2 == "作業予定"

        username_list = [e[0] for e in sum_worktime_df.columns if is_plan_column(e)]

        for username in username_list:
            # SettingWithCopyWarning を避けるため、暫定的に値をコピーする
            sum_worktime_df[(username, "作業予定_記号")] = sum_worktime_df[(username, "作業予定")].map(create_mark)

        output_columns = [("date", ""), ("dayofweek", "")] + [(e, "作業予定_記号") for e in username_list]
        sum_worktime_df[output_columns].to_csv(str(output_dir / "ユーザごとの作業予定_記号.csv"), encoding="utf_8_sig", index=False)

    @staticmethod
    @catch_exception
    def write_worktime_list(worktime_df: pandas.DataFrame, output_dir: Path, add_monitored_worktime: bool = False):
        worktime_df = worktime_df.rename(
            columns={
                "worktime_plan_hour": "作業予定時間",
                "worktime_result_hour": "作業実績時間",
                "worktime_monitored_hour": "計測時間",
                "working_description": "備考",
            }
        )
        columns = [
            "date",
            "organization_name",
            "project_title",
            "project_id",
            "username",
            "biography",
            "user_id",
            "作業予定時間",
            "作業実績時間",
            "計測時間",
            "備考",
        ]
        if not add_monitored_worktime:
            columns.remove("計測時間")

        worktime_df[columns].round(3).to_csv(str(output_dir / "作業時間の詳細一覧.csv"), encoding="utf_8_sig", index=False)

    @staticmethod
    @catch_exception
    def write_worktime_per_user_date(worktime_df_per_date_user: pandas.DataFrame, output_dir: Path):
        add_availabaility = "availability_hour" in worktime_df_per_date_user.columns
        target_renamed_columns = {"worktime_plan_hour": "作業予定時間", "worktime_result_hour": "作業実績時間"}
        if add_availabaility:
            target_renamed_columns.update({"availability_hour": "予定稼働時間"})

        df = worktime_df_per_date_user.rename(columns=target_renamed_columns)
        columns = [
            "date",
            "user_id",
            "username",
            "biography",
            "予定稼働時間",
            "作業予定時間",
            "作業実績時間",
        ]
        if not add_availabaility:
            columns.remove("予定稼働時間")

        df[columns].round(3).to_csv(str(output_dir / "日ごとの作業時間の一覧.csv"), encoding="utf_8_sig", index=False)

    @staticmethod
    @catch_exception
    def write_worktime_per_user(worktime_df_per_user: pandas.DataFrame, output_dir: Path, add_availability: bool):
        target_renamed_columns = {
            "worktime_plan_hour": "作業予定時間",
            "worktime_result_hour": "作業実績時間",
            "result_working_days": "実績稼働日数",
        }
        if add_availability:
            target_renamed_columns.update({"availability_hour": "予定稼働時間"})
            target_renamed_columns.update({"availability_days": "予定稼働日数"})

        df = worktime_df_per_user.rename(columns=target_renamed_columns)
        columns = [
            "user_id",
            "username",
            "biography",
            "予定稼働時間",
            "作業予定時間",
            "作業実績時間",
            "予定稼働日数",
            "実績稼働日数",
        ]
        if not add_availability:
            columns.remove("予定稼働時間")
            columns.remove("予定稼働日数")

        df[columns].round(3).to_csv(str(output_dir / "summary.csv"), encoding="utf_8_sig", index=False)

    def get_organization_member_list(
        self, organization_name_list: Optional[List[str]], project_id_list: Optional[List[str]]
    ) -> List[OrganizationMember]:
        member_list: List[OrganizationMember] = []

        if project_id_list is not None:
            tmp_organization_name_list = []
            for project_id in project_id_list:
                organization, _ = self.service.api.get_organization_of_project(project_id)
                tmp_organization_name_list.append(organization["organization_name"])

            organization_name_list = list(set(tmp_organization_name_list))

        if organization_name_list is not None:
            for organization_name in organization_name_list:
                member_list.extend(self.service.wrapper.get_all_organization_members(organization_name))

        return member_list

    def _get_user_from_organization_name_list(self, organization_name_list: List[str], user_id: str) -> Optional[User]:
        for organization_name in organization_name_list:
            organization_member = self.service.wrapper.get_organization_member_or_none(organization_name, user_id)
            if organization_member is not None:
                return User(
                    user_id=organization_member["user_id"],
                    account_id=organization_member["account_id"],
                    username=organization_member["username"],
                    biography=organization_member["biography"],
                )
        return None

    def _get_user_from_project_id_list(self, project_id_list: List[str], user_id: str) -> Optional[User]:
        for project_id in project_id_list:
            project_member = self.service.wrapper.get_project_member_or_none(project_id, user_id)
            if project_member is not None:
                return User(
                    user_id=project_member["user_id"],
                    account_id=project_member["account_id"],
                    username=project_member["username"],
                    biography=project_member["biography"],
                )
        return None

    def get_user_list(
        self,
        labor_list: List[LaborWorktime],
        organization_name_list: Optional[List[str]] = None,
        project_id_list: Optional[List[str]] = None,
        user_id_list: Optional[List[str]] = None,
    ) -> List[User]:
        """
        summary.csvに出力するユーザ一覧を取得する。

        Args:
            labor_list:
            organization_name_list:
            project_id_list:
            user_id_list:

        Returns:
            ユーザ一覧

        """
        df = (
            pandas.DataFrame(labor_list, columns=["account_id", "user_id", "username", "biography"])
            .drop_duplicates()
            .set_index("user_id")
        )

        user_list: List[User] = []

        if user_id_list is None:
            for user_id, row in df.iterrows():
                user_list.append(
                    User(
                        user_id=str(user_id),
                        account_id=row["account_id"],
                        username=row["username"],
                        biography=row["biography"],
                    )
                )
            return user_list

        if organization_name_list is not None:
            for user_id in user_id_list:
                if user_id in df.index:
                    row = df.loc[user_id]
                    user_list.append(
                        User(
                            user_id=str(user_id),
                            account_id=row["account_id"],
                            username=row["username"],
                            biography=row["biography"],
                        )
                    )
                    continue

                user = self._get_user_from_organization_name_list(organization_name_list, user_id)
                if user is not None:
                    user_list.append(user)
                else:
                    logger.warning(f"user_id={user_id} のユーザは、組織メンバに存在しませんでした。")

        if project_id_list is not None:
            for user_id in user_id_list:
                if user_id in df.index:
                    row = df.loc[user_id]
                    user_list.append(
                        User(
                            user_id=str(user_id),
                            account_id=row["account_id"],
                            username=row["username"],
                            biography=row["biography"],
                        )
                    )
                    continue

                user = self._get_user_from_project_id_list(project_id_list, user_id)
                if user is not None:
                    user_list.append(user)
                else:
                    logger.warning(f"user_id={user_id} のユーザは、プロジェクトメンバに存在しませんでした。")

        return user_list

    @staticmethod
    def get_availability_list(
        labor_availability_list: List[LaborAvailability],
        start_date: str,
        end_date: str,
    ) -> List[Optional[float]]:
        def get_availability_hour(str_date: str) -> Optional[float]:
            labor = more_itertools.first_true(labor_availability_list, pred=lambda e: e.date == str_date)
            if labor is not None:
                return labor.availability_hour
            else:
                return None

        availability_list: List[Optional[float]] = []
        for date in pandas.date_range(start=start_date, end=end_date):
            str_date = date.strftime(ListWorktimeByUserMain.DATE_FORMAT)
            availability_list.append(get_availability_hour(str_date))

        return availability_list

    def get_account_id_list_from_project_id(self, user_id_list: List[str], project_id_list: List[str]) -> List[str]:
        """
        project_idのリストから、対象ユーザのaccount_id を取得する。

        Args:
            user_id_list:
            organization_name_list:

        Returns:
            account_idのリスト
        """
        account_id_list = []
        not_exists_user_id_list = []
        for user_id in user_id_list:
            member_exists = False
            for project_id in project_id_list:
                member = self.facade.get_project_member_from_user_id(project_id, user_id)
                if member is not None:
                    account_id_list.append(member["account_id"])
                    member_exists = True
                    break
            if not member_exists:
                not_exists_user_id_list.append(user_id)

        if len(not_exists_user_id_list) == 0:
            return account_id_list
        else:
            raise ValueError(f"以下のユーザは、指定されたプロジェクトのプロジェクトメンバではありませんでした。\n{not_exists_user_id_list}")

    def get_account_id_list_from_organization_name(
        self, user_id_list: List[str], organization_name_list: List[str]
    ) -> List[str]:
        """
        組織名のリストから、対象ユーザのaccount_id を取得する。

        Args:
            user_id_list:
            organization_name_list:

        Returns:
            account_idのリスト
        """

        def _get_account_id(fuser_id: str) -> Optional[str]:
            # 脱退した可能性のあるユーザの組織メンバ情報を取得する
            for organization_name in organization_name_list:
                member = self.service.wrapper.get_organization_member_or_none(organization_name, fuser_id)
                if member is not None:
                    return member["account_id"]
            return None

        # 組織メンバの一覧をする（ただし脱退したメンバは取得できない）
        all_organization_member_list = []
        for organization_name in organization_name_list:
            all_organization_member_list.extend(self.service.wrapper.get_all_organization_members(organization_name))

        user_id_dict = {e["user_id"]: e["account_id"] for e in all_organization_member_list}
        account_id_list = []
        not_exists_user_id_list = []
        for user_id in user_id_list:
            if user_id in user_id_dict:
                account_id_list.append(user_id_dict[user_id])
            else:
                account_id = _get_account_id(user_id)
                if account_id is not None:
                    account_id_list.append(account_id)
                else:
                    not_exists_user_id_list.append(user_id)

        if len(not_exists_user_id_list) == 0:
            return account_id_list
        else:
            raise ValueError(f"以下のユーザは、指定された組織の組織メンバではありませんでした。\n{not_exists_user_id_list}")

    def get_labor_list(
        self,
        organization_name_list: Optional[List[str]],
        project_id_list: Optional[List[str]],
        user_id_list: Optional[List[str]],
        start_date: Optional[str],
        end_date: Optional[str],
        add_monitored_worktime: bool = False,
    ) -> List[LaborWorktime]:

        labor_list: List[LaborWorktime] = []
        account_id_list: Optional[List[str]] = None

        logger.info(f"労務管理情報を取得します。")
        if project_id_list is not None:
            if user_id_list is not None:
                account_id_list = self.get_account_id_list_from_project_id(
                    user_id_list, project_id_list=project_id_list
                )

            for project_id in project_id_list:
                labor_list.extend(
                    self.get_labor_list_from_project_id(
                        project_id,
                        account_id_list=account_id_list,
                        start_date=start_date,
                        end_date=end_date,
                        add_monitored_worktime=add_monitored_worktime,
                    )
                )

        elif organization_name_list is not None:
            if user_id_list is not None:
                account_id_list = self.get_account_id_list_from_organization_name(
                    user_id_list, organization_name_list=organization_name_list
                )
            for organization_name in organization_name_list:
                labor_list.extend(
                    self.get_labor_list_from_organization_name(
                        organization_name,
                        account_id_list=account_id_list,
                        start_date=start_date,
                        end_date=end_date,
                        add_monitored_worktime=add_monitored_worktime,
                    )
                )

        else:
            raise RuntimeError(f"organization_name_list or project_id_list を指定してください。")

        # 集計対象ユーザで絞り込む
        if user_id_list is not None:
            return [e for e in labor_list if e.user_id in user_id_list]
        else:
            return labor_list

    @staticmethod
    def create_sum_worktime_df(
        labor_list: List[LaborWorktime],
        user_list: List[User],
        start_date: str,
        end_date: str,
        labor_availability_list_dict: Optional[Dict[str, List[LaborAvailability]]] = None,
    ):
        reform_dict = {
            ("date", ""): [
                e.strftime(ListWorktimeByUserMain.DATE_FORMAT)
                for e in pandas.date_range(start=start_date, end=end_date)
            ],
            ("dayofweek", ""): [e.strftime("%a") for e in pandas.date_range(start=start_date, end=end_date)],
        }

        username_list = []
        for user in user_list:
            sum_worktime_list = ListWorktimeByUserMain.get_sum_worktime_list(
                labor_list, user_id=user.user_id, start_date=start_date, end_date=end_date
            )

            username_list.append(user.username)
            reform_dict.update(
                {
                    (user.username, "作業予定"): [e.worktime_plan_hour for e in sum_worktime_list],
                    (user.username, "作業実績"): [e.worktime_result_hour for e in sum_worktime_list],
                }
            )

            if labor_availability_list_dict is not None:
                labor_availability_list = labor_availability_list_dict.get(user.user_id, [])
                reform_dict.update(
                    {
                        (user.username, "予定稼働"): ListWorktimeByUserMain.get_availability_list(
                            labor_availability_list, start_date, end_date
                        )
                    }
                )

        key_list = ["作業予定", "作業実績", "予定稼働"] if labor_availability_list_dict else ["作業予定", "作業実績"]
        for key in key_list:
            data = numpy.array([reform_dict[(username, key)] for username in username_list], dtype=float)
            data = numpy.nan_to_num(data)
            reform_dict[("合計", key)] = list(numpy.sum(data, axis=0))

        columns = (
            [("date", ""), ("dayofweek", "")]
            + [("合計", key) for key in key_list]
            + [(username, key) for username in username_list for key in key_list]
        )

        sum_worktime_df = pandas.DataFrame(reform_dict, columns=columns)
        return sum_worktime_df

    @staticmethod
    def create_worktime_df_per_date_user(
        worktime_df: pandas.DataFrame,
        user_df: pandas.DataFrame,
        labor_availability_list_dict: Optional[Dict[str, List[LaborAvailability]]] = None,
    ) -> pandas.DataFrame:
        value_df = (
            worktime_df.pivot_table(
                values=["worktime_plan_hour", "worktime_result_hour"], index=["date", "user_id"], aggfunc=numpy.sum
            )
            .fillna(0)
            .reset_index()
        )
        if len(value_df) == 0:
            value_df = pandas.DataFrame(columns=["date", "user_id", "worktime_plan_hour", "worktime_result_hour"])

        if labor_availability_list_dict is not None:
            all_availability_list = []
            for availability_list in labor_availability_list_dict.values():
                all_availability_list.extend(availability_list)

            if len(all_availability_list) > 0:
                availability_df = pandas.DataFrame([e.to_dict() for e in all_availability_list])
            else:
                availability_df = pandas.DataFrame(columns=["date", "user_id", "availability_hour"])

            value_df = (
                value_df.merge(
                    availability_df[["date", "user_id", "availability_hour"]], how="outer", on=["date", "user_id"]
                )
                .fillna(0)
                .reset_index()
            )

        if len(value_df) > 0:
            return user_df.reset_index().merge(value_df, how="left", on=["user_id"]).reset_index()
        else:
            return pandas.DataFrame()

    @staticmethod
    def set_day_count_to_dataframe(
        worktime_df_per_date_user: pandas.DataFrame, value_df: pandas.DataFrame, worktime_column: str, days_column: str
    ):
        df_filter = worktime_df_per_date_user[worktime_df_per_date_user[worktime_column] > 0]
        if len(df_filter) > 0:
            value_df[days_column] = (
                df_filter.pivot_table(index=["user_id"], values="worktime_result_hour", aggfunc="count")
                .fillna(0)
                .astype(pandas.Int64Dtype())
            )
        else:
            value_df[days_column] = 0

    @staticmethod
    def get_value_columns(columns: pandas.Series) -> pandas.Series:
        """
        ユーザ情報以外のcolumnsを取得する
        """
        return columns.drop(["user_id", "username", "biography"])

    @staticmethod
    def create_worktime_df_per_user(
        worktime_df_per_date_user: pandas.DataFrame, user_df: pandas.DataFrame, add_availability: bool = False
    ) -> pandas.DataFrame:
        if len(worktime_df_per_date_user) > 0:
            value_df = worktime_df_per_date_user.pivot_table(index=["user_id"], aggfunc=numpy.sum).fillna(0)
            ListWorktimeByUserMain.set_day_count_to_dataframe(
                worktime_df_per_date_user,
                value_df,
                worktime_column="worktime_result_hour",
                days_column="result_working_days",
            )
            ListWorktimeByUserMain.set_day_count_to_dataframe(
                worktime_df_per_date_user,
                value_df,
                worktime_column="worktime_plan_hour",
                days_column="plan_working_days",
            )

            if add_availability:
                ListWorktimeByUserMain.set_day_count_to_dataframe(
                    worktime_df_per_date_user,
                    value_df,
                    worktime_column="availability_hour",
                    days_column="availability_days",
                )
            value_df.fillna(0, inplace=True)

        else:
            columns = [
                "availability_hour",
                "worktime_plan_hour",
                "worktime_result_hour",
                "availability_days",
                "plan_working_days",
                "result_working_days",
            ]
            if not add_availability:
                columns.remove("availability_hour")
                columns.remove("availability_days")
            value_df = pandas.DataFrame(columns=columns)

        user_df.set_index("user_id", inplace=True)

        df = user_df.join(value_df, how="left").reset_index()
        value_columns = ListWorktimeByUserMain.get_value_columns(df.columns)
        df[value_columns] = df[value_columns].fillna(0)
        return df

    def write_labor_list(
        self,
        labor_list: List[LaborWorktime],
        user_list: List[User],
        start_date: str,
        end_date: str,
        output_dir: Path,
        labor_availability_list_dict: Optional[Dict[str, List[LaborAvailability]]] = None,
        add_monitored_worktime: bool = False,
    ):
        if len(labor_list) > 0:
            worktime_df = pandas.DataFrame([e.to_dict() for e in labor_list])
            self.write_worktime_list(worktime_df, output_dir, add_monitored_worktime)
        else:
            worktime_df = pandas.DataFrame(columns=["date", "user_id", "worktime_plan_hour", "worktime_result_hour"])
            logger.info("予定作業時間または実績作業時間が入力されているデータは見つからなかったので、'作業時間の詳細一覧.csv' は出力しません。")

        # 日ごとの作業時間の一覧.csv を出力
        user_df = pandas.DataFrame(user_list)
        worktime_df_per_date_user = self.create_worktime_df_per_date_user(
            worktime_df=worktime_df, user_df=user_df, labor_availability_list_dict=labor_availability_list_dict
        )
        if len(worktime_df_per_date_user) > 0:
            self.write_worktime_per_user_date(worktime_df_per_date_user, output_dir)
        else:
            logger.info("日ごとの作業時間情報に関するデータが見つからなかったので、 '日ごとの作業時間の一覧.csv' は出力しません。")

        add_availability = labor_availability_list_dict is not None
        worktime_df_per_user = self.create_worktime_df_per_user(
            worktime_df_per_date_user=worktime_df_per_date_user, user_df=user_df, add_availability=add_availability
        )
        self.write_worktime_per_user(worktime_df_per_user, output_dir, add_availability=add_availability)

        # 行方向に日付、列方向にユーザを表示したDataFrame
        sum_worktime_df = self.create_sum_worktime_df(
            labor_list=labor_list,
            user_list=user_list,
            start_date=start_date,
            end_date=end_date,
            labor_availability_list_dict=labor_availability_list_dict,
        )
        self.write_sum_worktime_list(sum_worktime_df, output_dir)
        self.write_sum_plan_worktime_list(sum_worktime_df, output_dir)

    def main(
        self,
        organization_name_list: Optional[List[str]],
        project_id_list: Optional[List[str]],
        user_id_list: Optional[List[str]],
        start_date: Optional[str],
        end_date: Optional[str],
        output_dir: Path,
        add_availability: bool = False,
        add_monitored_worktime: bool = False,
    ) -> None:
        """
        作業時間の一覧を出力する
        """
        labor_list = self.get_labor_list(
            organization_name_list=organization_name_list,
            project_id_list=project_id_list,
            user_id_list=user_id_list,
            start_date=start_date,
            end_date=end_date,
            add_monitored_worktime=add_monitored_worktime,
        )

        if len(labor_list) == 0:
            logger.info(f"予定/実績に関する労務管理情報が0件です。")
            if start_date is None or end_date is None or user_id_list is None:
                logger.info(f"後続の処理を続けることができないので終了します。")
                return

        if start_date is None or end_date is None:
            sorted_date_list = sorted({e.date for e in labor_list if e.worktime_result_hour > 0})
            if start_date is None:
                start_date = sorted_date_list[0]
            if end_date is None:
                end_date = sorted_date_list[-1]

        logger.info(f"集計期間: start_date={start_date}, end_date={end_date}")
        user_list = self.get_user_list(
            labor_list,
            organization_name_list=organization_name_list,
            project_id_list=project_id_list,
            user_id_list=user_id_list,
        )
        logger.info(f"集計対象ユーザの数: {len(user_list)}")

        labor_availability_list_dict: Optional[Dict[str, List[LaborAvailability]]] = None
        if add_availability:
            labor_availability_list_dict = self.get_labor_availability_list_dict(
                user_list=user_list,
                start_date=start_date,
                end_date=end_date,
            )

        self.write_labor_list(
            labor_list=labor_list,
            user_list=user_list,
            start_date=start_date,
            end_date=end_date,
            output_dir=output_dir,
            labor_availability_list_dict=labor_availability_list_dict,
            add_monitored_worktime=add_monitored_worktime,
        )

    def get_user_id_list_from_project_id_list(self, project_id_list: List[str]) -> List[str]:
        """
        プロジェクトメンバ一覧からuser_id_listを取得する。
        Args:
            project_id_list:

        Returns:
            user_id_list

        """
        member_list: List[Dict[str, Any]] = []
        for project_id in project_id_list:
            member_list.extend(self.service.wrapper.get_all_project_members(project_id))
        user_id_list = [e["user_id"] for e in member_list]
        return list(set(user_id_list))

    @staticmethod
    def get_first_and_last_date(str_month: str) -> Tuple[str, str]:
        """
        年月（"YYYY-MM"）から、月初と月末の日付を返す。

        Args:
            str_month: 年月（"YYYY-MM"）

        Returns:
            月初と月末の日付が格納されたタプル

        """
        dt_first_date = datetime.datetime.strptime(str_month, ListWorktimeByUserMain.MONTH_FORMAT)
        _, days = calendar.monthrange(dt_first_date.year, dt_first_date.month)
        dt_last_date = dt_first_date + datetime.timedelta(days=(days - 1))
        return (
            dt_first_date.strftime(ListWorktimeByUserMain.DATE_FORMAT),
            dt_last_date.strftime(ListWorktimeByUserMain.DATE_FORMAT),
        )

    @staticmethod
    def get_start_and_end_date_from_month(start_month: str, end_month: str) -> Tuple[str, str]:
        """
        開始月、終了月から、開始日付、終了日付を取得する。

        Args:
            start_month:
            end_month:

        Returns:


        """
        first_date, _ = ListWorktimeByUserMain.get_first_and_last_date(start_month)
        _, end_date = ListWorktimeByUserMain.get_first_and_last_date(end_month)
        return first_date, end_date


class ListWorktimeByUser(AbstractCommandLineInterface):
    """
    作業時間をユーザごとに出力する。
    """

    @staticmethod
    def get_start_and_end_date_from_args(args: argparse.Namespace) -> Tuple[Optional[str], Optional[str]]:
        if args.start_date is not None:
            start_date = args.start_date
        elif args.start_month is not None:
            start_date, _ = ListWorktimeByUserMain.get_first_and_last_date(args.start_month)
        else:
            start_date = None

        if args.end_date is not None:
            end_date = args.end_date
        elif args.end_month is not None:
            _, end_date = ListWorktimeByUserMain.get_first_and_last_date(args.end_month)
        else:
            end_date = None

        return (start_date, end_date)

    def main(self) -> None:
        args = self.args

        arg_user_id_list = get_list_from_args(args.user_id) if args.user_id is not None else None
        project_id_list = get_list_from_args(args.project_id) if args.project_id is not None else None
        organization_name_list = get_list_from_args(args.organization) if args.organization is not None else None

        start_date, end_date = self.get_start_and_end_date_from_args(args)
        output_dir = Path(args.output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)

        mian_obj = ListWorktimeByUserMain(self.service)
        mian_obj.main(
            organization_name_list=organization_name_list,
            project_id_list=project_id_list,
            start_date=start_date,
            end_date=end_date,
            output_dir=output_dir,
            user_id_list=arg_user_id_list,
            add_availability=args.add_availability,
            add_monitored_worktime=args.add_monitored_worktime,
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListWorktimeByUser(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    target_group = parser.add_mutually_exclusive_group(required=True)
    target_group.add_argument(
        "-org",
        "--organization",
        type=str,
        nargs="+",
        help="集計対象の組織名を指定してください。`file://`を先頭に付けると、組織名の一覧が記載されたファイルを指定できます。",
    )

    target_group.add_argument(
        "-p",
        "--project_id",
        type=str,
        nargs="+",
        help="集計対象のプロジェクトを指定してください。`file://`を先頭に付けると、project_idの一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument(
        "-u",
        "--user_id",
        type=str,
        nargs="+",
        help="集計対象のユーザのuser_idを指定してください。`--organization`を指定した場合は必須です。"
        "指定しない場合は、プロジェクトメンバが指定されます。"
        "`file://`を先頭に付けると、user_idの一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument("--add_availability", action="store_true", help="指定した場合、'ユーザごとの作業時間.csv'に予定稼働時間も出力します。")
    parser.add_argument(
        "--add_monitored_worktime", action="store_true", help="指定した場合、'作業時間の詳細一覧.csv'にAnnoFab計測時間も出力します。"
    )

    start_period_group = parser.add_mutually_exclusive_group()
    start_period_group.add_argument("--start_date", type=str, help="集計期間の開始日(YYYY-MM-DD)")
    start_period_group.add_argument("--start_month", type=str, help="集計期間の開始月(YYYY-MM-DD)")

    end_period_group = parser.add_mutually_exclusive_group()
    end_period_group.add_argument("--end_date", type=str, help="集計期間の終了日(YYYY-MM)")
    end_period_group.add_argument("--end_month", type=str, help="集計期間の終了月(YYYY-MM)")

    parser.add_argument("-o", "--output_dir", type=str, required=True, help="出力先のディレクトリのパス")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_worktime_by_user"
    subcommand_help = "ユーザごとに作業予定時間、作業実績時間を出力します。"
    description = "ユーザごとに作業予定時間、作業実績時間を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

import more_itertools
import numpy
import pandas
import requests
from annofabapi.models import ProjectMemberRole, ProjectMemberStatus
from dataclasses_json import DataClassJsonMixin

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login

logger = logging.getLogger(__name__)


@dataclass
class Member(DataClassJsonMixin):
    """
    登録するプロジェクトメンバ
    """

    user_id: str
    member_role: ProjectMemberRole
    sampling_inspection_rate: Optional[int]
    sampling_acceptance_rate: Optional[int]


class PutProjectMembers(AbstractCommandLineInterface):
    """
    プロジェクトメンバをCSVで登録する。
    """

    @staticmethod
    def find_member(members: List[Dict[str, Any]], user_id: str) -> Optional[Dict[str, Any]]:
        member = more_itertools.first_true(members, default=None, pred=lambda e: e["user_id"] == user_id)
        return member

    @staticmethod
    def member_exists(members: List[Dict[str, Any]], user_id) -> bool:
        return PutProjectMembers.find_member(members, user_id) is not None

    def invite_project_member(self, project_id, member: Member, old_project_members: List[Dict[str, Any]]):
        old_member = self.find_member(old_project_members, member.user_id)
        last_updated_datetime = old_member["updated_datetime"] if old_member is not None else None

        request_body = {
            "member_status": ProjectMemberStatus.ACTIVE.value,
            "member_role": member.member_role.value,
            "sampling_inspection_rate": member.sampling_inspection_rate,
            "sampling_acceptance_rate": member.sampling_acceptance_rate,
            "last_updated_datetime": last_updated_datetime,
        }
        updated_project_member = self.service.api.put_project_member(
            project_id, member.user_id, request_body=request_body
        )[0]
        return updated_project_member

    def delete_project_member(self, project_id, deleted_member: Dict[str, Any]):
        request_body = {
            "member_status": ProjectMemberStatus.INACTIVE.value,
            "member_role": deleted_member["member_role"],
            "last_updated_datetime": deleted_member["updated_datetime"],
        }
        updated_project_member = self.service.api.put_project_member(
            project_id, deleted_member["user_id"], request_body=request_body
        )[0]
        return updated_project_member

    def put_project_members(self, project_id: str, members: List[Member], delete: bool = False):
        """
        プロジェクトメンバを一括で登録する。

        Args:
            project_id: プロジェクトメンバの登録先のプロジェクトのプロジェクトID
            members: 登録するプロジェクトメンバのList
            delete: Trueならば、mebersにないメンバを、対象プロジェクトから削除する。

        """

        super().validate_project(project_id, [ProjectMemberRole.OWNER])

        organization_name = self.facade.get_organization_name_from_project_id(project_id)
        organization_members = self.service.wrapper.get_all_organization_members(organization_name)

        old_project_members = self.service.wrapper.get_all_project_members(project_id)
        project_title = self.facade.get_project_title(project_id)

        count_invite_members = 0
        # プロジェクトメンバを登録
        logger.info(f"{project_title} に、{len(members)} 件のプロジェクトメンバを登録します。")
        for member in members:
            if member.user_id == self.service.api.login_user_id:
                logger.debug(f"ユーザ '{member.user_id}'は自分自身なので、登録しません。")
                continue

            if not self.member_exists(organization_members, member.user_id):
                logger.warning(f"ユーザ '{member.user_id}' は、" f"'{organization_name}' 組織の組織メンバでないため、登録できませんでした。")
                continue

            message_for_confirm = (
                f"ユーザ '{member.user_id}'を、{project_title} プロジェクトのメンバに登録しますか？" f"member_role={member.member_role.value}"
            )
            if not self.confirm_processing(message_for_confirm):
                continue

            # メンバを登録
            try:
                self.invite_project_member(project_id, member, old_project_members)
                logger.debug(
                    f"user_id = {member.user_id}, member_role = {member.member_role.value} のユーザをプ" f"ロジェクトメンバに登録しました。"
                )
                count_invite_members += 1

            except requests.exceptions.HTTPError as e:
                logger.warning(e)
                logger.warning(
                    f"プロジェクトメンバの登録に失敗しました。" f"user_id = {member.user_id}, member_role = {member.member_role.value}"
                )

        logger.info(f"{project_title} に、{count_invite_members} / {len(members)} 件のプロジェクトメンバを登録しました。")

        # プロジェクトメンバを削除
        if delete:
            user_id_list = [e.user_id for e in members]
            # 自分自身は削除しないようにする
            deleted_members = [
                e
                for e in old_project_members
                if (e["user_id"] not in user_id_list and e["user_id"] != self.service.api.login_user_id)
            ]

            count_delete_members = 0
            logger.info(f"{project_title} から、{len(deleted_members)} 件のプロジェクトメンバを削除します。")
            for deleted_member in deleted_members:
                message_for_confirm = f"ユーザ '{deleted_member['user_id']}'を、" f"{project_title} のプロジェクトメンバから削除しますか？"
                if not self.confirm_processing(message_for_confirm):
                    continue

                try:
                    self.delete_project_member(project_id, deleted_member)
                    logger.debug(f"ユーザ '{deleted_member['user_id']}' をプロジェクトメンバから削除しました。")
                    count_delete_members += 1
                except requests.exceptions.HTTPError as e:
                    logger.warning(e)
                    logger.warning(f"プロジェクトメンバの削除に失敗しました。user_id = '{deleted_member['user_id']}' ")

            logger.info(f"{project_title} から {count_delete_members} / {len(deleted_members)} 件の" f"プロジェクトメンバを削除しました。")

    @staticmethod
    def get_members_from_csv(csv_path: Path) -> List[Member]:
        def create_member(e):
            return Member(
                user_id=e.user_id,
                member_role=ProjectMemberRole(e.member_role),
                sampling_inspection_rate=e.sampling_inspection_rate,
                sampling_acceptance_rate=e.sampling_acceptance_rate,
            )

        df = pandas.read_csv(
            str(csv_path),
            sep=",",
            header=None,
            names=("user_id", "member_role", "sampling_inspection_rate", "sampling_acceptance_rate"),
        ).replace({numpy.nan: None})
        members = [create_member(e) for e in df.itertuples()]
        return members

    def main(self):
        args = self.args
        members = self.get_members_from_csv(Path(args.csv))
        self.put_project_members(args.project_id, members=members, delete=args.delete)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    PutProjectMembers(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    parser.add_argument(
        "--csv",
        type=str,
        required=True,
        help=(
            "プロジェクトメンバが記載されたCVファイルのパスを指定してください。"
            "CSVのフォーマットは、「1列目:user_id(required), 2列目:member_role(required), "
            "3列目:sampling_inspection_rate, 4列目:sampling_acceptance_rate, ヘッダ行なし, カンマ区切り」です。"
            "member_roleは `owner`, `worker`, `accepter`, `training_data_user` のいずれかです。"
            "sampling_inspection_rate, sampling_acceptance_rate を省略した場合は未設定になります。"
            "ただし自分自身は登録しません。"
        ),
    )

    parser.add_argument("--delete", action="store_true", help="CSVファイルに記載されていないプロジェクトメンバを削除します。ただし自分自身は削除しません。")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "put"
    subcommand_help = "プロジェクトメンバを登録する。"
    description = "プロジェクトメンバを登録する。"
    epilog = "オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import json
import logging
from enum import Enum
from typing import List

import pandas
from annofabapi.models import ProjectMemberRole, Task, TaskPhase, TaskStatus

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_json_from_args,
    get_wait_options_from_args,
)
from annofabcli.common.dataclasses import WaitOptions
from annofabcli.common.download import DownloadingFile
from annofabcli.common.enums import FormatArgument
from annofabcli.statistics.summarize_task_count import get_step_for_current_phase

logger = logging.getLogger(__name__)

DEFAULT_WAIT_OPTIONS = WaitOptions(interval=60, max_tries=360)
DEFAULT_TASK_ID_DELIMITER = "_"


class TaskStatusForSummary(Enum):
    """
    TaskStatusのサマリー用（知りたい情報をstatusにしている）
    """

    COMPLETE = "complete"
    ON_HOLD = "on_hold"
    ANNOTATION_NOT_STARTED = "annotation_not_started"
    """教師付作業されていない状態"""
    INSPECTION_NOT_STARTED = "inspection_not_started"
    """検査作業されていない状態"""
    ACCEPTANCE_NOT_STARTED = "acceptance_not_started"
    """受入作業されていない状態"""
    OTHER = "other"
    """休憩中/作業中/2回目移行の各フェーズの未着手状態"""

    @staticmethod
    def _get_not_started_status(task: Task) -> "TaskStatusForSummary":
        # `number_of_inspections=1`を指定する理由：多段検査を無視して、検査フェーズが１回目かどうかを知りたいため
        step = get_step_for_current_phase(task, number_of_inspections=1)
        if step == 1:
            phase = TaskPhase(task["phase"])
            if phase == TaskPhase.ANNOTATION:
                return TaskStatusForSummary.ANNOTATION_NOT_STARTED
            elif phase == TaskPhase.INSPECTION:
                return TaskStatusForSummary.INSPECTION_NOT_STARTED
            elif phase == TaskPhase.ACCEPTANCE:
                return TaskStatusForSummary.ACCEPTANCE_NOT_STARTED
            else:
                raise ValueError(f"'{phase.value}'は対象外です。")
        else:
            return TaskStatusForSummary.OTHER

    @staticmethod
    def from_task(task: Task) -> "TaskStatusForSummary":
        status = TaskStatus(task["status"])
        if status == TaskStatus.COMPLETE:
            return TaskStatusForSummary.COMPLETE

        elif status == TaskStatus.ON_HOLD:
            return TaskStatusForSummary.ON_HOLD

        elif status == TaskStatus.NOT_STARTED:
            return TaskStatusForSummary._get_not_started_status(task)
        else:
            # WORKING, BREAK
            return TaskStatusForSummary.OTHER


def get_task_id_prefix(task_id: str, delimiter: str = DEFAULT_TASK_ID_DELIMITER):
    tmp_list = task_id.split(delimiter)
    if len(tmp_list) <= 1:
        return "unknown"
    else:
        return delimiter.join(tmp_list[0 : len(tmp_list) - 1])


def add_task_id_prefix_to_task_list(task_list: List[Task], delimiter: str = DEFAULT_TASK_ID_DELIMITER) -> List[Task]:
    for task in task_list:
        task["task_id_prefix"] = get_task_id_prefix(task["task_id"], delimiter=delimiter)
        task["status_for_summary"] = TaskStatusForSummary.from_task(task).value

    return task_list


def create_task_count_summary_df(task_list: List[Task], delimiter: str = DEFAULT_TASK_ID_DELIMITER) -> pandas.DataFrame:
    """
    タスク数を集計したDataFrameを生成する。

    Args:
        task_list:

    Returns:

    """

    def add_columns_if_not_exists(df: pandas.DataFrame, column: str):
        if column not in df.columns:
            df[column] = 0

    df_task = pandas.DataFrame(add_task_id_prefix_to_task_list(task_list, delimiter=delimiter))

    df_summary = df_task.pivot_table(
        values="task_id", index=["task_id_prefix"], columns=["status_for_summary"], aggfunc="count", fill_value=0
    ).reset_index()

    for status in TaskStatusForSummary:
        add_columns_if_not_exists(df_summary, status.value)

    df_summary["sum"] = (
        df_task.pivot_table(values="task_id", index=["task_id_prefix"], aggfunc="count", fill_value=0)
        .reset_index()
        .fillna(0)["task_id"]
    )

    return df_summary


class SummarizeTaskCountByTaskId(AbstractCommandLineInterface):
    def print_summarize_task_count(self, df: pandas.DataFrame) -> None:
        columns = ["task_id_prefix"] + [status.value for status in TaskStatusForSummary] + ["sum"]
        annofabcli.utils.print_according_to_format(
            df[columns],
            arg_format=FormatArgument(FormatArgument.CSV),
            output=self.output,
            csv_format=self.csv_format,
        )

    def main(self):
        args = self.args
        project_id = args.project_id
        super().validate_project(project_id, [ProjectMemberRole.OWNER, ProjectMemberRole.TRAINING_DATA_USER])

        if args.task_json is not None:
            task_json_path = args.task_json
        else:
            wait_options = get_wait_options_from_args(get_json_from_args(args.wait_options), DEFAULT_WAIT_OPTIONS)
            cache_dir = annofabcli.utils.get_cache_dir()
            task_json_path = cache_dir / f"{project_id}-task.json"

            downloading_obj = DownloadingFile(self.service)
            downloading_obj.download_task_json(
                project_id, dest_path=str(task_json_path), is_latest=args.latest, wait_options=wait_options
            )

        with open(task_json_path, encoding="utf-8") as f:
            task_list = json.load(f)

        df = create_task_count_summary_df(task_list, delimiter=args.delimiter)
        self.print_summarize_task_count(df)


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    parser.add_argument(
        "--task_json",
        type=str,
        help="タスク情報が記載されたJSONファイルのパスを指定してます。JSONファイルは`$ annofabcli project download task`コマンドで取得できます。"
        "指定しない場合は、AnnoFabからタスク全件ファイルをダウンロードします。",
    )

    parser.add_argument(
        "--delimiter",
        type=str,
        default="_",
        help="task_idのprefixと連番を分ける区切り文字です。デフォルトは`_`で、`{prefix}_{連番}`のようなtask_idを想定しています。"
        "指定しない場合は、AnnoFabからタスク全件ファイルをダウンロードします。",
    )

    parser.add_argument(
        "--latest", action="store_true", help="最新のタスク一覧ファイルを参照します。このオプションを指定すると、タスク一覧ファイルを更新するのに数分待ちます。"
    )

    parser.add_argument(
        "--wait_options",
        type=str,
        help="タスク一覧ファイルの更新が完了するまで待つ際のオプションを、JSON形式で指定してください。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルは`{"interval":60, "max_tries":360}` です。'
        "`interval`:完了したかを問い合わせる間隔[秒], "
        "`max_tires`:完了したかの問い合わせを最大何回行うか。",
    )

    argument_parser.add_csv_format()
    argument_parser.add_output()

    parser.set_defaults(subcommand_func=main)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    SummarizeTaskCountByTaskId(service, facade, args).main()


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "summarize_task_count_by_task_id"
    subcommand_help = "task_idのプレフィックスごとに、タスク数を出力します。"
    description = "task_idのプレフィックスごとに、タスク数をCSV形式で出力します。"
    epilog = "アノテーションユーザまたはオーナロールを持つユーザで実行してください。"
    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description=description, epilog=epilog
    )
    parse_args(parser)



import argparse
import asyncio
import json
import logging
import zipfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import annofabapi
import pandas
from annofabapi.parser import SimpleAnnotationZipParser

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.download import DownloadingFile

logger = logging.getLogger(__name__)


def _millisecond_to_hour(millisecond: int):
    return millisecond / 1000 / 3600


def _get_time_range(str_data: str):
    tmp_list = str_data.split(",")
    return (int(tmp_list[0]), int(tmp_list[1]))


class ListOutOfRangeAnnotationForMovieMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service

    @staticmethod
    def get_max_seconds_for_webapi(annotation: Dict[str, Any]) -> Tuple[float, float]:
        details = annotation["details"]
        range_list = [_get_time_range(e["data"]) for e in details if e["data"] is not None]
        if len(range_list) == 0:
            return 0, 0
        else:
            max_begin = max([e[0] for e in range_list]) / 1000
            max_end = max([e[1] for e in range_list]) / 1000
            return max_begin, max_end

    @staticmethod
    def get_max_seconds_for_zip(annotation: Dict[str, Any]) -> Tuple[float, float]:
        details = annotation["details"]
        range_list = [(e["data"]["begin"], e["data"]["end"]) for e in details if e["data"]["_type"] == "Range"]
        if len(range_list) == 0:
            return 0, 0
        else:
            max_begin = max([e[0] for e in range_list]) / 1000
            max_end = max([e[1] for e in range_list]) / 1000
            return max_begin, max_end

    def create_dataframe(
        self,
        proejct_id: str,
        task_list: List[Dict[str, Any]],
        input_data_list: List[Dict[str, Any]],
        annotation_zip: Optional[Path],
    ) -> pandas.DataFrame:
        if annotation_zip is None:
            logger.info(f"{len(task_list)} 件のアノテーション情報をWebAPIで取得します。")
            for task_index, task in enumerate(task_list):
                task["worktime_hour"] = _millisecond_to_hour(task["work_time_span"])
                task["input_data_id"] = task["input_data_id_list"][0]
                annotation, _ = self.service.api.get_editor_annotation(
                    proejct_id, task["task_id"], task["input_data_id"]
                )
                max_seconds = self.get_max_seconds_for_webapi(annotation)
                task["max_begin_second"] = max_seconds[0]
                task["max_end_second"] = max_seconds[1]
                if (task_index + 1) % 100 == 0:
                    logger.info(f"{task_index+1} 件のアノテーション情報を取得しました。")
        else:
            logger.info(f"{len(task_list)} 件のアノテーション情報を {str(annotation_zip)} から取得します。")
            with zipfile.ZipFile(str(annotation_zip), "r") as zip_file:
                for task_index, task in enumerate(task_list):
                    task["worktime_hour"] = _millisecond_to_hour(task["work_time_span"])
                    task["input_data_id"] = task["input_data_id_list"][0]

                    parser = SimpleAnnotationZipParser(zip_file, f"{task['task_id']}/{task['input_data_id']}.json")
                    simple_annotation = parser.load_json()
                    max_seconds = self.get_max_seconds_for_zip(simple_annotation)
                    task["max_begin_second"] = max_seconds[0]
                    task["max_end_second"] = max_seconds[1]

                    if (task_index + 1) % 100 == 0:
                        logger.info(f"{task_index+1} 件のアノテーション情報を取得しました。")

        df_task = pandas.DataFrame(
            task_list,
            columns=[
                "task_id",
                "status",
                "phase",
                "worktime_hour",
                "max_begin_second",
                "max_end_second",
                "input_data_id",
            ],
        )
        df_input_data = pandas.DataFrame(input_data_list, columns=["input_data_id", "input_duration"])
        df_merged = pandas.merge(df_task, df_input_data, how="left", on="input_data_id")
        return df_merged

    @staticmethod
    def filter_task_list(task_list: List[Dict[str, Any]], task_id_list: List[str]) -> List[Dict[str, Any]]:
        def _exists(task_id) -> bool:
            if task_id in task_id_set:
                task_id_set.remove(task_id)
                return True
            else:
                return False

        task_id_set = set(task_id_list)
        task_list = [e for e in task_list if _exists(e["task_id"])]
        if len(task_id_set) > 0:
            tmp = "\n".join(task_id_set)
            logger.warning(f"以下のタスクは存在しません。\n{tmp}")
        return task_list

    def list_out_of_range_annotation_for_movie(
        self, project_id: str, task_id_list: Optional[List[str]], parse_annotation_zip: bool = False
    ) -> pandas.DataFrame:
        cache_dir = annofabcli.utils.get_cache_dir()
        downloading_obj = DownloadingFile(self.service)

        input_data_json_path = cache_dir / f"input_data-{project_id}.json"
        task_json_path = cache_dir / f"task-{project_id}.json"
        awaitable_list = [
            downloading_obj.download_task_json_with_async(project_id, dest_path=str(task_json_path)),
            downloading_obj.download_input_data_json_with_async(
                project_id,
                dest_path=str(input_data_json_path),
            ),
        ]

        annotation_zip_path = None
        if parse_annotation_zip:
            annotation_zip_path = cache_dir / f"annotation-{project_id}.zip"
            awaitable_list.append(
                downloading_obj.download_annotation_zip_with_async(
                    project_id,
                    dest_path=str(annotation_zip_path),
                )
            )

        gather = asyncio.gather(*awaitable_list)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(gather)

        with input_data_json_path.open() as f:
            input_data_list = json.load(f)

        with task_json_path.open() as f:
            task_list = json.load(f)
            if task_id_list is not None:
                task_list = self.filter_task_list(task_list, task_id_list)

        df = self.create_dataframe(
            project_id, task_list=task_list, input_data_list=input_data_list, annotation_zip=annotation_zip_path
        )
        return df


class ListOutOfRangeAnnotationForMovie(AbstractCommandLineInterface):
    def main(self):
        args = self.args

        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id) if args.task_id is not None else None
        main_obj = ListOutOfRangeAnnotationForMovieMain(self.service)
        df = main_obj.list_out_of_range_annotation_for_movie(args.project_id, task_id_list, args.parse_annotation_zip)
        self.print_csv(df)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListOutOfRangeAnnotationForMovie(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)
    argument_parser.add_project_id()
    argument_parser.add_task_id(required=False)
    argument_parser.add_output()

    parser.add_argument(
        "--parse_annotation_zip", action="store_true", help="アノテーションzipから範囲外のアノテーション情報を取得します。ただし、範囲外の終了時間は存在しません。"
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_out_of_range_annotation_for_movie"
    subcommand_help = "動画範囲外のアノテーションを探すためのCSVを出力します。"
    description = "動画範囲外のアノテーションを探すためのCSVを出力します。最後尾のアノテーションの開始時間、終了時間を出力します。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import asyncio
import json
import logging
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional

import annofabapi
import pandas
from annofabapi.models import ProjectMemberRole

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_json_from_args,
    get_list_from_args,
    get_wait_options_from_args,
)
from annofabcli.common.dataclasses import WaitOptions
from annofabcli.common.download import DownloadingFile
from annofabcli.common.enums import FormatArgument

logger = logging.getLogger(__name__)

DEFAULT_WAIT_OPTIONS = WaitOptions(interval=60, max_tries=360)


def millisecond_to_hour(millisecond: int):
    return millisecond / 1000 / 3600


class ListInputDataMergedTaskMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    @staticmethod
    def _to_task_list_based_input_data(task_list: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        new_all_task_list: List[Dict[str, Any]] = []
        for task in task_list:
            for input_data_id in task["input_data_id_list"]:
                new_task = {
                    "task_id": task["task_id"],
                    "status": task["status"],
                    "phase": task["phase"],
                    "worktime_hour": millisecond_to_hour(task["work_time_span"]),
                    "input_data_id": input_data_id,
                }
                new_all_task_list.append(new_task)
        return new_all_task_list

    @staticmethod
    def _filter_input_data(
        df_input_data: pandas.DataFrame,
        input_data_id_list: Optional[List[str]] = None,
        input_data_name_list: Optional[List[str]] = None,
    ):
        df = df_input_data
        if input_data_id_list is not None:
            df = df[df["input_data_id"].isin(input_data_id_list)]

        if input_data_name_list is not None:
            df = pandas.concat(
                [
                    df[df["input_data_name"].str.lower().str.contains(input_data_name.lower())]
                    for input_data_name in input_data_name_list
                ]
            )
        return df

    def create_input_data_merged_task(
        self,
        input_data_list: List[Dict[str, Any]],
        task_list: List[Dict[str, Any]],
        input_data_id_list: Optional[List[str]] = None,
        input_data_name_list: Optional[List[str]] = None,
    ):
        new_task_list = self._to_task_list_based_input_data(task_list)

        df_input_data = pandas.DataFrame(input_data_list)
        df_input_data = self._filter_input_data(
            df_input_data, input_data_id_list=input_data_id_list, input_data_name_list=input_data_name_list
        )

        df_task = pandas.DataFrame(new_task_list)

        df_merged = pandas.merge(df_input_data, df_task, how="left", on="input_data_id")

        return df_merged


class ListInputDataMergedTask(AbstractCommandLineInterface):
    def __init__(self, service: annofabapi.Resource, facade: AnnofabApiFacade, args: argparse.Namespace):
        super().__init__(service, facade, args)

    @staticmethod
    def validate(args: argparse.Namespace) -> bool:
        COMMON_MESSAGE = "annofabcli input_data list_merged_task: error:"
        if args.project_id is None and (args.input_data_json is None or args.task_json is None):
            print(
                f"{COMMON_MESSAGE} '--project_id' か、'--task_json'/'--input_data_json'ペアのいずれかを指定する必要があります。",
                file=sys.stderr,
            )
            return False

        if (args.input_data_json is None and args.task_json is not None) or (
            args.input_data_json is not None and args.task_json is None
        ):
            print(
                f"{COMMON_MESSAGE} '--task_json'と'--input_data_json'の両方を指定する必要があります。",
                file=sys.stderr,
            )
            return False

        return True

    def download_json_files(self, project_id: str, output_dir: Path, is_latest: bool, wait_options: WaitOptions):
        loop = asyncio.get_event_loop()
        downloading_obj = DownloadingFile(self.service)
        gather = asyncio.gather(
            downloading_obj.download_input_data_json_with_async(
                project_id,
                dest_path=str(output_dir / "input_data.json"),
                is_latest=is_latest,
                wait_options=wait_options,
            ),
            downloading_obj.download_task_json_with_async(
                project_id, dest_path=str(output_dir / "task.json"), is_latest=is_latest, wait_options=wait_options
            ),
        )
        loop.run_until_complete(gather)

    def main(self):
        args = self.args
        if not self.validate(args):
            return

        project_id = args.project_id
        if project_id is not None:
            super().validate_project(project_id, [ProjectMemberRole.OWNER, ProjectMemberRole.TRAINING_DATA_USER])
            wait_options = get_wait_options_from_args(get_json_from_args(args.wait_options), DEFAULT_WAIT_OPTIONS)
            cache_dir = annofabcli.utils.get_cache_dir()
            self.download_json_files(project_id, cache_dir, args.latest, wait_options)
            task_json_path = cache_dir / "task.json"
            input_data_json_path = cache_dir / "input_data.json"
        else:
            task_json_path = args.task_json
            input_data_json_path = args.input_data_json

        with open(task_json_path, encoding="utf-8") as f:
            task_list = json.load(f)

        with open(input_data_json_path, encoding="utf-8") as f:
            input_data_list = json.load(f)

        input_data_id_list = get_list_from_args(args.input_data_id) if args.input_data_id is not None else None
        input_data_name_list = get_list_from_args(args.input_data_name) if args.input_data_name is not None else None

        main_obj = ListInputDataMergedTaskMain(self.service)
        df_merged = main_obj.create_input_data_merged_task(
            input_data_list=input_data_list,
            task_list=task_list,
            input_data_id_list=input_data_id_list,
            input_data_name_list=input_data_name_list,
        )
        annofabcli.utils.print_according_to_format(
            df_merged, arg_format=FormatArgument(FormatArgument.CSV), output=self.output, csv_format=self.csv_format
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListInputDataMergedTask(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    parser.add_argument(
        "-p", "--project_id", type=str, help="対象のプロジェクトのproject_idを指定してください。" "指定すると、入力データ一覧ファイル、タスク一覧ファイルをダウンロードします。"
    )

    parser.add_argument(
        "--input_data_json",
        type=str,
        help="入力データ情報が記載されたJSONファイルのパスを指定してください。JSONに記載された情報を元に出力します。"
        "JSONファイルは`$ annofabcli project download input_data`コマンドで取得できます。",
    )

    parser.add_argument(
        "--task_json",
        type=str,
        help="タスク情報が記載されたJSONファイルのパスを指定してください。JSONに記載された情報を元に出力します。"
        "JSONファイルは`$ annofabcli project download task`コマンドで取得できます。",
    )

    parser.add_argument(
        "--latest",
        action="store_true",
        help="入力データ一覧ファイル、タスク一覧ファイルの更新が完了するまで待って、最新のファイルをダウンロードします。" "'--project_id'を指定したときのみ有効です。",
    )

    parser.add_argument(
        "--wait_options",
        type=str,
        help="入力データ一覧ファイル、タスク一覧ファイルの更新が完了するまで待つ際のオプションを、JSON形式で指定してください。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルは`{"interval":60, "max_tries":360}` です。'
        "`interval`:完了したかを問い合わせる間隔[秒], "
        "`max_tires`:完了したかの問い合わせを最大何回行うか。",
    )

    parser.add_argument("-i", "--input_data_id", type=str, nargs="+", help="指定したinput_data_idに完全一致する入力データを絞り込みます。")
    parser.add_argument(
        "--input_data_name", type=str, nargs="+", help="指定したinput_data_nameに部分一致(大文字小文字区別しない）する入力データを絞り込みます。"
    )

    argument_parser.add_output()
    argument_parser.add_csv_format()

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_merged_task"
    subcommand_help = "タスク一覧と結合した入力データ一覧のCSVを出力します。"
    description = "タスク一覧と結合した入力データ一覧のCSVを出力します。"
    epilog = "アノテーションユーザまたはオーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import datetime
import itertools
import json
import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

import annofabapi
import dateutil
import pandas
from annofabapi.models import ProjectMemberRole, Task
from annofabapi.utils import str_now
from dataclasses_json import DataClassJsonMixin
from dateutil.parser import parse
from more_itertools import first_true

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_json_from_args,
    get_wait_options_from_args,
)
from annofabcli.common.dataclasses import WaitOptions
from annofabcli.common.download import DownloadingFile
from annofabcli.common.enums import FormatArgument
from annofabcli.common.utils import isoduration_to_hour
from annofabcli.statistics.summarize_task_count_by_task_id import TaskStatusForSummary

logger = logging.getLogger(__name__)

DEFAULT_WAIT_OPTIONS = WaitOptions(interval=60, max_tries=360)


@dataclass
class TaskPhaseStatistics(DataClassJsonMixin):
    date: str
    annotation_worktime: float = 0
    inspection_worktime: float = 0
    acceptance_worktime: float = 0

    def sum_worktime(self) -> float:
        return self.annotation_worktime + self.inspection_worktime + self.acceptance_worktime


@dataclass
class MonitoredWorktime(DataClassJsonMixin):
    """AnnoFab計測時間の詳細情報"""

    sum: float
    """合計時間[hour]"""
    annotation: float
    inspection: float
    acceptance: float


@dataclass
class RemainingTaskCount(DataClassJsonMixin):
    """
    フェーズごとのタスク数
    """

    complete: int = 0
    """完了状態のタスク数"""

    annotation_not_started: int = 0
    """一度も教師付作業されていないタスク数"""

    inspection_not_started: int = 0
    """一度も検査作業されていないタスク数"""

    acceptance_not_started: int = 0
    """一度も受入作業されていないタスク数"""

    on_hold: int = 0
    """保留状態のタスク数"""

    other: int = 0
    """休憩中/作業中/2回目以降の各フェーズの未着手状態のタスク数"""


@dataclass
class ProgressData(DataClassJsonMixin):
    # task_count: TaskCount
    # """タスク数情報"""

    actual_worktime: float
    """実績作業時間[hour]"""

    task_count: int
    """完了したタスク数"""

    input_data_count: int
    """完了したタスク配下の入力データ数"""

    velocity_per_task: Optional[float]
    """acutal_worktime/task_count"""

    velocity_per_input_data: Optional[float]
    """acutal_worktime/input_data_count"""

    monitored_worktime: Optional[float] = None
    """AnnoFabの計測作業時間[hour]"""

    annotation_monitored_worktime: Optional[float] = None
    """AnnoFabのannotation計測作業時間[hour]"""

    inspection_monitored_worktime: Optional[float] = None
    """AnnoFabのinspection計測作業時間[hour]"""

    acceptance_monitored_worktime: Optional[float] = None
    """AnnoFabのacceptance計測作業時間[hour]"""


@dataclass
class ResultValues(DataClassJsonMixin):

    cumulation: ProgressData
    """累計情報"""
    today: ProgressData
    """対象日当日の情報"""
    week: ProgressData
    """対象日から1週間（直前7日間）の情報"""


@dataclass
class Planvalues(DataClassJsonMixin):
    plan_worktime: float
    """予定作業時間"""
    task_count: Optional[int]
    """予定の完了タスク数"""


@dataclass
class DashboardData(DataClassJsonMixin):
    project_id: str
    project_title: str
    date: str
    """対象日（YYYY-MM-DD）"""
    measurement_datetime: str
    """計測日時。（2004-04-01T12:00+09:00形式）"""

    remaining_task_count: RemainingTaskCount
    """残りのタスク数"""
    result: ResultValues
    plan: Dict[int, Planvalues]
    task_exhaustation_date: Optional[str]
    """タスク枯渇予定日"""


def add_info_to_task(task: Task):
    task["status_for_summary"] = TaskStatusForSummary.from_task(task).value


def get_remaining_task_count_info_from_task_list(task_list: List[Task]) -> RemainingTaskCount:
    """
    状態ごとのタスク数を取得する。

    Args:
        task_list:

    Returns:

    """
    status_list: List[str] = [e["status_for_summary"] for e in task_list]
    tmp = pandas.Series.value_counts(status_list)
    task_count = RemainingTaskCount.from_dict(tmp.to_dict())
    return task_count


def get_completed_task_count_and_input_data_count(task_list: List[Task]) -> Tuple[int, int]:
    """
    完了したタスク数と入力データ数を求める

    Args:
        task_list:

    Returns:
        Tuple[task_count, input_data_count]
    """
    task_list = [e for e in task_list if e["status_for_summary"] == TaskStatusForSummary.COMPLETE.value]
    input_data_count = len(list(itertools.chain.from_iterable([e["input_data_id_list"] for e in task_list])))
    task_count = len(task_list)
    return task_count, input_data_count


def _to_datetime_from_date(date: datetime.date) -> datetime.datetime:
    dt = datetime.datetime.combine(date, datetime.datetime.min.time())
    return dt.replace(tzinfo=dateutil.tz.tzlocal())


def millisecond_to_hour(millisecond: int):
    return millisecond / 1000 / 3600


def get_task_list_where_updated_datetime(
    task_list: List[Task], lower_date: datetime.date, upper_date: datetime.date
) -> List[Task]:
    """
    タスクの更新日が、対象期間内であるタスク一覧を取得する。

    Args:
        task_list:
        lower_date:
        upper_date:

    Returns:

    """

    def pred(task):
        updated_datetime = parse(task["updated_datetime"])
        return lower_datetime <= updated_datetime < upper_datetime

    lower_datetime = _to_datetime_from_date(lower_date)
    upper_datetime = _to_datetime_from_date(upper_date + datetime.timedelta(days=1))
    return [t for t in task_list if pred(t)]


def _get_actual_worktime_hour_from_labor(labor: Dict[str, Any]) -> float:
    working_time_by_user = labor["values"]["working_time_by_user"]
    if working_time_by_user is None:
        return 0

    actual_worktime = working_time_by_user.get("results")
    if actual_worktime is None:
        return 0
    else:
        return actual_worktime / 3600 / 1000


def _get_plan_worktime_hour_from_labor(labor: Dict[str, Any]) -> float:
    working_time_by_user = labor["values"]["working_time_by_user"]
    if working_time_by_user is None:
        return 0

    actual_worktime = working_time_by_user.get("plans")
    if actual_worktime is None:
        return 0
    else:
        return actual_worktime / 3600 / 1000


def _get_today_info(
    today: str,
    task_list: List[Task],
    actual_worktime_dict: Dict[str, float],
    task_phase_statistics: List[TaskPhaseStatistics],
) -> ProgressData:
    dt_today = datetime.datetime.strptime(today, "%Y-%m-%d").date()
    task_list_for_day = get_task_list_where_updated_datetime(task_list, lower_date=dt_today, upper_date=dt_today)
    today_monitor_worktime_info = get_monitored_worktime(
        task_phase_statistics, lower_date=dt_today, upper_date=dt_today
    )
    task_count, input_data_count = get_completed_task_count_and_input_data_count(task_list_for_day)
    actual_worktime = actual_worktime_dict.get(today, 0)
    today_info = ProgressData(
        task_count=task_count,
        input_data_count=input_data_count,
        actual_worktime=actual_worktime,
        velocity_per_task=actual_worktime / task_count if task_count > 0 else None,
        velocity_per_input_data=actual_worktime / input_data_count if input_data_count > 0 else None,
    )
    if today_monitor_worktime_info is not None:
        today_info.monitored_worktime = today_monitor_worktime_info.sum
        today_info.annotation_monitored_worktime = today_monitor_worktime_info.annotation
        today_info.inspection_monitored_worktime = today_monitor_worktime_info.inspection
        today_info.acceptance_monitored_worktime = today_monitor_worktime_info.acceptance
    return today_info


def _get_week_info(
    today: str,
    task_list: List[Task],
    actual_worktime_dict: Dict[str, float],
    task_phase_statistics: List[TaskPhaseStatistics],
) -> ProgressData:
    dt_today = datetime.datetime.strptime(today, "%Y-%m-%d").date()
    week_ago = dt_today - datetime.timedelta(days=6)
    task_list_for_week = get_task_list_where_updated_datetime(task_list, lower_date=week_ago, upper_date=dt_today)
    week_monitor_worktime_info = get_monitored_worktime(task_phase_statistics, lower_date=week_ago, upper_date=dt_today)
    task_count, input_data_count = get_completed_task_count_and_input_data_count(task_list_for_week)
    actual_worktime = get_worktime_for_period(actual_worktime_dict, lower_date=week_ago, upper_date=dt_today)

    seven_days_info = ProgressData(
        task_count=task_count,
        input_data_count=input_data_count,
        actual_worktime=actual_worktime,
        velocity_per_task=actual_worktime / task_count if task_count > 0 else None,
        velocity_per_input_data=actual_worktime / input_data_count if input_data_count > 0 else None,
    )
    if week_monitor_worktime_info is not None:
        seven_days_info.monitored_worktime = week_monitor_worktime_info.sum
        seven_days_info.annotation_monitored_worktime = week_monitor_worktime_info.annotation
        seven_days_info.inspection_monitored_worktime = week_monitor_worktime_info.inspection
        seven_days_info.acceptance_monitored_worktime = week_monitor_worktime_info.acceptance
    return seven_days_info


def get_worktime_for_period(worktime_dict: Dict[str, float], lower_date: datetime.date, upper_date: datetime.date):
    sum_worktime = 0.0
    for dt in pandas.date_range(start=lower_date, end=upper_date):
        str_date = str(dt.date())
        sum_worktime += worktime_dict.get(str_date, 0.0)
    return sum_worktime


def get_monitored_worktime(
    task_phase_statistics: List[TaskPhaseStatistics], lower_date: datetime.date, upper_date: datetime.date
) -> Optional[MonitoredWorktime]:
    upper_stat = first_true(task_phase_statistics, pred=lambda e: e.date == str(upper_date))
    lower_stat = first_true(
        task_phase_statistics, pred=lambda e: e.date == str(lower_date - datetime.timedelta(days=1))
    )
    if upper_stat is None or lower_stat is None:
        logger.debug(f"{lower_date} 〜 {upper_date} 期間のmonitor_worktimeを算出できませんでした。")
        return None

    return MonitoredWorktime(
        sum=upper_stat.sum_worktime() - lower_stat.sum_worktime(),
        annotation=upper_stat.annotation_worktime - lower_stat.annotation_worktime,
        inspection=upper_stat.inspection_worktime - lower_stat.inspection_worktime,
        acceptance=upper_stat.acceptance_worktime - lower_stat.acceptance_worktime,
    )


def _get_cumulation_info(task_list: List[Task], actual_worktime_dict: Dict[str, float]) -> ProgressData:
    task_count, input_data_count = get_completed_task_count_and_input_data_count(task_list)
    actual_worktime = sum(actual_worktime_dict.values())

    return ProgressData(
        task_count=task_count,
        input_data_count=input_data_count,
        monitored_worktime=sum([t["worktime_hour"] for t in task_list]),
        actual_worktime=actual_worktime,
        velocity_per_task=actual_worktime / task_count if task_count > 0 else None,
        velocity_per_input_data=actual_worktime / input_data_count if input_data_count > 0 else None,
    )


def _get_plan_value(
    plan_worktime_dict: Dict[str, float],
    dt_today: datetime.date,
    days: int,
    velocity_per_task: Optional[float],
) -> Planvalues:
    dt_from_date = dt_today + datetime.timedelta(days=1)
    dt_end_date = dt_today + datetime.timedelta(days=days)

    plan_worktime = get_worktime_for_period(plan_worktime_dict, lower_date=dt_from_date, upper_date=dt_end_date)
    task_count = (
        int(plan_worktime / velocity_per_task) if velocity_per_task is not None and velocity_per_task > 0 else None
    )
    return Planvalues(plan_worktime=plan_worktime, task_count=task_count)


def get_task_exhaustation_date(
    plan_worktime_dict: Dict[str, float], annotation_not_started: int, velocity_per_task: float
) -> Optional[str]:
    """
    タスクの枯渇予定日を求める。
    教師付されていないタスク数がすでに０ならば、Noneを返す。

    Args:
        plan_worktime_dict:
        annotation_not_started:
        velocity_per_task:

    Returns:

    """
    if annotation_not_started <= 0 or velocity_per_task <= 0:
        return None

    remaining_task: float = annotation_not_started
    for date in sorted(plan_worktime_dict.keys()):
        plan_worktime = plan_worktime_dict[date]
        task_count = plan_worktime / velocity_per_task
        remaining_task -= task_count
        if remaining_task <= 0:
            return date
    return None


def create_plan_values(
    plan_worktime_dict: Dict[str, float], velocity_per_task: Optional[float], today: str
) -> Dict[int, Planvalues]:
    dt_today = datetime.datetime.strptime(today, "%Y-%m-%d").date()
    return {
        7: _get_plan_value(plan_worktime_dict, dt_today=dt_today, days=7, velocity_per_task=velocity_per_task),
        14: _get_plan_value(plan_worktime_dict, dt_today=dt_today, days=14, velocity_per_task=velocity_per_task),
    }


class PrintDashBoardMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    def get_task_phase_statistics(self, project_id: str) -> List[TaskPhaseStatistics]:
        """
        フェーズごとの累積作業時間をCSVに出力するための dict 配列を作成する。

        Args:
            project_id:

        Returns:
            フェーズごとの累積作業時間に対応するdict配列

        """
        task_phase_statistics = self.service.wrapper.get_task_phase_statistics(project_id)
        row_list: List[TaskPhaseStatistics] = []
        for stat_by_date in task_phase_statistics:
            elm = {"date": stat_by_date["date"]}
            for phase_stat in stat_by_date["phases"]:
                phase = phase_stat["phase"]
                worktime_hour = isoduration_to_hour(phase_stat["worktime"])
                elm[f"{phase}_worktime"] = worktime_hour
            row_list.append(TaskPhaseStatistics.from_dict(elm))
        return row_list

    def get_actual_worktime_dict(self, project_id: str, date: str) -> Dict[str, float]:
        """
        日毎のプロジェクト全体の実績作業時間を取得する。

        Args:
            project_id:
            date: 対象期間の終了日

        Returns:
            key:date, value:実績作業時間のdict

        """
        # 予定稼働時間を取得するには、特殊な組織IDを渡す
        labor_list, _ = self.service.api.get_labor_control({"project_id": project_id, "to": date})
        actual_worktime_dict: Dict[str, float] = defaultdict(float)
        for labor in labor_list:
            date = labor["date"]
            actual_worktime = _get_actual_worktime_hour_from_labor(labor)
            actual_worktime_dict[date] += actual_worktime

        return actual_worktime_dict

    def get_plan_worktime_dict(self, project_id: str, today: str, days: Optional[int] = None) -> Dict[str, float]:
        """
        日毎のプロジェクト全体の予定作業時間を取得する。

        Args:
            project_id:
            date: 対象期間の終了日

        Returns:
            key:date, value: 予定作業時間のdict

        """
        dt_today = datetime.datetime.strptime(today, "%Y-%m-%d").date()
        dt_from_date = dt_today + datetime.timedelta(days=1)
        end_date = str(dt_today + datetime.timedelta(days=days)) if days is not None else None

        # 予定稼働時間を取得するには、特殊な組織IDを渡す
        labor_list, _ = self.service.api.get_labor_control(
            {"project_id": project_id, "from": str(dt_from_date), "to": end_date}
        )
        plan_worktime_dict: Dict[str, float] = defaultdict(float)
        for labor in labor_list:
            date = labor["date"]
            plan_worktime = _get_plan_worktime_hour_from_labor(labor)
            plan_worktime_dict[date] += plan_worktime

        return plan_worktime_dict

    def create_result_values(self, project_id: str, date: str, task_list: List[Task]) -> ResultValues:
        """
        実績情報を生成する。

        Args:
            task_list:
            date: タスク数　集計対象日

        Returns:

        """
        for task in task_list:
            task["status_for_summary"] = TaskStatusForSummary.from_task(task).value
            task["worktime_hour"] = millisecond_to_hour(task["work_time_span"])

        actual_worktime_dict = self.get_actual_worktime_dict(project_id, date)
        task_phase_statistics = self.get_task_phase_statistics(project_id)

        cumulation_info = _get_cumulation_info(task_list, actual_worktime_dict)
        today_info = _get_today_info(
            today=date,
            task_list=task_list,
            actual_worktime_dict=actual_worktime_dict,
            task_phase_statistics=task_phase_statistics,
        )
        week_info = _get_week_info(
            today=date,
            task_list=task_list,
            actual_worktime_dict=actual_worktime_dict,
            task_phase_statistics=task_phase_statistics,
        )
        return ResultValues(cumulation=cumulation_info, today=today_info, week=week_info)

    def create_dashboard_data(self, project_id: str, date: str, task_list: List[Task]) -> DashboardData:
        project_title = self.facade.get_project_title(project_id)
        result = self.create_result_values(project_id=project_id, date=date, task_list=task_list)
        remaining_task_count = get_remaining_task_count_info_from_task_list(task_list)
        plan_worktime_dict = self.get_plan_worktime_dict(project_id, today=date)
        velocity_per_task = result.week.velocity_per_task

        task_exhaustation_date = (
            get_task_exhaustation_date(
                plan_worktime_dict=plan_worktime_dict,
                annotation_not_started=remaining_task_count.annotation_not_started,
                velocity_per_task=velocity_per_task,
            )
            if velocity_per_task is not None
            else None
        )

        dashboard_info = DashboardData(
            project_id=project_id,
            project_title=project_title,
            date=date,
            measurement_datetime=str_now(),
            remaining_task_count=remaining_task_count,
            result=result,
            plan=create_plan_values(plan_worktime_dict, velocity_per_task=velocity_per_task, today=date),
            task_exhaustation_date=task_exhaustation_date,
        )
        return dashboard_info


class DashBoard(AbstractCommandLineInterface):
    def print_summarize_task_count(self, target_dict: dict) -> None:
        annofabcli.utils.print_according_to_format(
            target_dict,
            arg_format=FormatArgument.JSON,
            output=self.output,
        )

    def main(self):
        args = self.args
        project_id = args.project_id
        super().validate_project(project_id, [ProjectMemberRole.OWNER, ProjectMemberRole.TRAINING_DATA_USER])

        if args.task_json is not None:
            task_json_path = args.task_json
        else:
            wait_options = get_wait_options_from_args(get_json_from_args(args.wait_options), DEFAULT_WAIT_OPTIONS)
            cache_dir = annofabcli.utils.get_cache_dir()
            task_json_path = cache_dir / f"{project_id}-task.json"

            downloading_obj = DownloadingFile(self.service)
            downloading_obj.download_task_json(
                project_id, dest_path=str(task_json_path), is_latest=args.latest, wait_options=wait_options
            )

        with open(task_json_path, encoding="utf-8") as f:
            task_list = json.load(f)

        main_obj = PrintDashBoardMain(self.service)

        dashboard_date = main_obj.create_dashboard_data(project_id, date=args.date, task_list=task_list)
        annofabcli.utils.print_according_to_format(
            dashboard_date.to_dict(),
            arg_format=FormatArgument(self.str_format),
            output=self.output,
        )


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    parser.add_argument(
        "--task_json",
        type=str,
        help="タスク情報が記載されたJSONファイルのパスを指定してます。JSONファイルは`$ annofabcli project download task`コマンドで取得できます。"
        "指定しない場合は、AnnoFabからタスク全件ファイルをダウンロードします。",
    )

    parser.add_argument("--date", type=str, required=True, help="`YYYY-MM-DD`形式で実績の対象日を指定してください。")

    parser.add_argument(
        "--latest", action="store_true", help="最新のタスク一覧ファイルを参照します。このオプションを指定すると、タスク一覧ファイルを更新するのに数分待ちます。"
    )

    parser.add_argument(
        "--wait_options",
        type=str,
        help="タスク一覧ファイルの更新が完了するまで待つ際のオプションを、JSON形式で指定してください。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルは`{"interval":60, "max_tries":360}` です。'
        "`interval`:完了したかを問い合わせる間隔[秒], "
        "`max_tires`:完了したかの問い合わせを最大何回行うか。",
    )

    argument_parser.add_format([FormatArgument.PRETTY_JSON, FormatArgument.JSON], default=FormatArgument.PRETTY_JSON)
    argument_parser.add_output()

    parser.set_defaults(subcommand_func=main)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    DashBoard(service, facade, args).main()


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "dashboard"
    subcommand_help = "ダッシュボードとなる情報（タスク数など）をJSON形式で出力します。"
    description = "ダッシュボードとなる情報（タスク数など）をJSON形式で出力します。"
    epilog = "オーナロールを持つユーザで実行してください。"
    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description=description, epilog=epilog
    )
    parse_args(parser)



import argparse

import annofabcli
import annofabcli.annotation.change_annotation_attributes
import annofabcli.annotation.delete_annotation
import annofabcli.annotation.dump_annotation
import annofabcli.annotation.import_annotation
import annofabcli.annotation.list_annotation_count
import annofabcli.annotation.restore_anotation
import annofabcli.common.cli


def parse_args(parser: argparse.ArgumentParser):

    subparsers = parser.add_subparsers(dest="subcommand_name")

    # サブコマンドの定義
    annofabcli.annotation.change_annotation_attributes.add_parser(subparsers)
    annofabcli.annotation.delete_annotation.add_parser(subparsers)
    annofabcli.annotation.dump_annotation.add_parser(subparsers)
    annofabcli.annotation.import_annotation.add_parser(subparsers)
    annofabcli.annotation.list_annotation_count.add_parser(subparsers)
    annofabcli.annotation.restore_anotation.add_parser(subparsers)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "annotation"
    subcommand_help = "アノテーション関係のサブコマンド"
    description = "アノテーション関係のサブコマンド"

    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description, is_subcommand=False
    )
    parse_args(parser)



import argparse
import copy
import json
import logging
from pathlib import Path
from typing import Any, Dict

from annofabapi.models import JobType, ProjectMemberRole

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_json_from_args,
    get_wait_options_from_args,
)
from annofabcli.common.dataclasses import WaitOptions

logger = logging.getLogger(__name__)

DEFAULT_WAIT_OPTIONS = WaitOptions(interval=60, max_tries=360)


class PutTask(AbstractCommandLineInterface):
    """
    CSVからタスクを登録する。
    """

    DEFAULT_BY_COUNT = {"allow_duplicate_input_data": False, "input_data_order": "name_asc"}

    def put_task_by_count(self, project_id: str, task_generate_rule: Dict[str, Any]):
        project_last_updated_datetime = self.service.api.get_project(project_id)[0]["updated_datetime"]
        task_generate_rule.update({"_type": "ByCount"})
        request_body = {
            "task_generate_rule": task_generate_rule,
            "project_last_updated_datetime": project_last_updated_datetime,
        }
        self.service.api.initiate_tasks_generation(project_id, request_body=request_body)

    def put_task_from_csv_file(self, project_id: str, csv_file: Path) -> None:
        """
        CSVファイルからタスクを登録する。

        Args:
            project_id:
            csv_file: タスク登録に関する情報が記載されたCSV
        """
        project_title = self.facade.get_project_title(project_id)
        logger.info(f"{project_title} に対して、{str(csv_file)} からタスクを登録します。")
        self.service.wrapper.initiate_tasks_generation_by_csv(project_id, csvfile_path=str(csv_file))

    def wait_for_completion(
        self,
        project_id: str,
        wait_options: WaitOptions,
        wait: bool = False,
    ) -> None:
        """
        CSVファイルからタスクを登録する。

        Args:
            project_id:
            wait_options: タスク登録の完了を待つ処理
            wait: タスク登録が完了するまで待つかどうか
        """
        logger.info(f"タスクの登録中です（サーバ側の処理）。")

        if wait:
            MAX_WAIT_MINUTUE = wait_options.max_tries * wait_options.interval / 60
            logger.info(f"最大{MAX_WAIT_MINUTUE}分間、タスク登録処理が終了するまで待ちます。")

            result = self.service.wrapper.wait_for_completion(
                project_id,
                job_type=JobType.GEN_TASKS,
                job_access_interval=wait_options.interval,
                max_job_access=wait_options.max_tries,
            )
            if result:
                logger.info(f"タスクの登録が完了しました。")
            else:
                logger.warning(f"タスクの登録に失敗しました。または、{MAX_WAIT_MINUTUE}分間待っても、タスクの登録が完了しませんでした。")

    def main(self):
        args = self.args
        project_id = args.project_id
        super().validate_project(project_id, [ProjectMemberRole.OWNER])

        if args.csv is not None:
            csv_file = Path(args.csv)
            self.put_task_from_csv_file(project_id, csv_file)
        elif args.by_count is not None:
            by_count = copy.deepcopy(PutTask.DEFAULT_BY_COUNT)
            by_count.update(get_json_from_args(args.by_count))
            self.put_task_by_count(project_id, by_count)
        else:
            raise RuntimeError("--csv or --by_count が指定されていません。")

        wait_options = get_wait_options_from_args(get_json_from_args(args.wait_options), DEFAULT_WAIT_OPTIONS)
        self.wait_for_completion(
            project_id,
            wait=args.wait,
            wait_options=wait_options,
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    PutTask(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    file_group = parser.add_mutually_exclusive_group(required=True)
    file_group.add_argument(
        "--csv",
        type=str,
        help=(
            "タスクに割り当てる入力データが記載されたCSVファイルのパスを指定してください。"
            "CSVのフォーマットは、「1列目:task_id,2列目:Any(無視される), 3列目:input_data_id」です。"
            "タスク作成画面でアップロードするCSVと同じフォーマットです。"
        ),
    )

    file_group.add_argument(
        "--by_count",
        type=str,
        help=f"1つのタスクに割り当てる入力データの個数などの情報を、JSON形式で指定してください。"
        "JSONフォーマットは https://annofab.com/docs/api/#operation/initiateTasksGeneration"
        " APIのリクエストボディ 'task_generate_rule'と同じです。"
        f"デフォルトは'{json.dumps(PutTask.DEFAULT_BY_COUNT)}'です。"
        "'file://'を先頭に付けるとjsonファイルを指定できます。",
    )

    parser.add_argument("--wait", action="store_true", help=("タスク登録が完了するまで待ちます。"))

    parser.add_argument(
        "--wait_options",
        type=str,
        help="タスクの登録が完了するまで待つ際のオプションを、JSON形式で指定してください。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルは`{"interval":60, "max_tries":360}` です。'
        "`interval`:完了したかを問い合わせる間隔[秒], "
        "`max_tires`:完了したかの問い合わせを最大何回行うか。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "put"
    subcommand_help = "タスクを登録します。"
    description = "タスクに割り当てる入力データが記載されたCSVから、タスクを登録します。"
    epilog = "オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import importlib
import json
import logging.handlers
import re
from dataclasses import dataclass
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

import annofabapi
import pandas
from annofabapi.models import ProjectMemberRole, TaskPhase
from dataclasses_json import DataClassJsonMixin

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login
from annofabcli.statistics.csv import FILENAME_WHOLE_PEFORMANCE, Csv, write_summarise_whole_peformance_csv
from annofabcli.statistics.database import Database, Query
from annofabcli.statistics.linegraph import LineGraph, OutputTarget
from annofabcli.statistics.merge_visualization_dir import merge_visualization_dir
from annofabcli.statistics.scatter import Scatter
from annofabcli.statistics.table import AggregationBy, Table

logger = logging.getLogger(__name__)


@dataclass
class CommnadLineArgs(DataClassJsonMixin):
    task_query: Optional[Dict[str, Any]]
    user_id_list: Optional[List[str]]
    start_date: Optional[str]
    end_date: Optional[str]
    ignored_task_id_list: Optional[List[str]]


@dataclass
class ProjectSummary(DataClassJsonMixin):
    project_id: str
    project_title: str
    measurement_datetime: str
    """計測日時。（2004-04-01T12:00+09:00形式）"""
    args: CommnadLineArgs
    """コマンドライン引数"""


def write_project_name_file(
    annofab_service: annofabapi.Resource, project_id: str, command_line_args: CommnadLineArgs, output_project_dir: Path
):
    """
    ファイル名がプロジェクト名のjsonファイルを生成する。
    """
    project_info = annofab_service.api.get_project(project_id)[0]
    project_title = project_info["title"]
    logger.info(f"project_titile = {project_title}")
    filename = annofabcli.utils.to_filename(project_title)
    output_project_dir.mkdir(exist_ok=True, parents=True)

    prject_summary = ProjectSummary(
        project_id=project_id,
        project_title=project_title,
        measurement_datetime=annofabapi.utils.str_now(),
        args=command_line_args,
    )

    with open(str(output_project_dir / f"{filename}.json"), "w") as f:
        json.dump(prject_summary.to_dict(), f, ensure_ascii=False, indent=2)


def catch_exception(function: Callable[..., Any]) -> Callable[..., Any]:
    """
    Exceptionをキャッチしてログにstacktraceを出力する。
    """

    def wrapped(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(e)
            logger.exception(e)

    return wrapped


def get_project_output_dir(project_title: str) -> str:
    return re.sub(r'[\\/:*?"<>|]+', "__", project_title)


class WriteCsvGraph:
    task_df: Optional[pandas.DataFrame] = None
    annotation_df: Optional[pandas.DataFrame] = None
    account_statistics_df: Optional[pandas.DataFrame] = None
    df_by_date_user_for_annotation: Optional[pandas.DataFrame] = None
    df_by_date_user_for_acceptance: Optional[pandas.DataFrame] = None
    df_by_date_user_for_inspection: Optional[pandas.DataFrame] = None
    task_history_df: Optional[pandas.DataFrame] = None
    labor_df: Optional[pandas.DataFrame] = None
    productivity_df: Optional[pandas.DataFrame] = None
    whole_productivity_df: Optional[pandas.DataFrame] = None

    def __init__(self, table_obj: Table, output_dir: Path, minimal_output: bool = False):
        self.table_obj = table_obj
        self.csv_obj = Csv(str(output_dir))
        # holivesのloadに時間がかかって、helpコマンドの出力が遅いため、遅延ロードする
        histogram_module = importlib.import_module("annofabcli.statistics.histogram")
        self.histogram_obj = histogram_module.Histogram(str(output_dir / "histogram"))  # type: ignore
        self.linegraph_obj = LineGraph(str(output_dir / "line-graph"))
        self.scatter_obj = Scatter(str(output_dir / "scatter"))
        self.minimal_output = minimal_output

    def _get_task_df(self):
        if self.task_df is None:
            self.task_df = self.table_obj.create_task_df()
        return self.task_df

    def _get_task_history_df(self):
        if self.task_history_df is None:
            self.task_history_df = self.table_obj.create_task_history_df()
        return self.task_history_df

    def _get_annotation_df(self):
        if self.annotation_df is None:
            self.annotation_df = self.table_obj.create_task_for_annotation_df()
        return self.annotation_df

    def _get_account_statistics_df(self):
        if self.account_statistics_df is None:
            self.account_statistics_df = self.table_obj.create_account_statistics_df()
        return self.account_statistics_df

    def _get_df_by_date_user_for_annotation(self):
        if self.df_by_date_user_for_annotation is None:
            task_df = self._get_task_df()
            self.df_by_date_user_for_annotation = self.table_obj.create_dataframe_by_date_user_for_annotation(task_df)
        return self.df_by_date_user_for_annotation

    def _get_df_by_date_user_for_inspection(self):
        if self.df_by_date_user_for_inspection is None:
            task_df = self._get_task_df()
            self.df_by_date_user_for_inspection = self.table_obj.create_dataframe_by_date_user_for_inspection(task_df)
        return self.df_by_date_user_for_inspection

    def _get_df_by_date_user_for_acceptance(self):
        if self.df_by_date_user_for_acceptance is None:
            task_df = self._get_task_df()
            self.df_by_date_user_for_acceptance = self.table_obj.create_dataframe_by_date_user_for_acceptance(task_df)
        return self.df_by_date_user_for_acceptance

    def _get_labor_df(self):
        if self.labor_df is None:
            self.labor_df = self.table_obj.create_labor_df()
        return self.labor_df

    def _get_productivity_df(self):
        if self.productivity_df is None:
            task_history_df = self._get_task_history_df()
            task_df = self._get_task_df()
            annotation_count_ratio_df = self.table_obj.create_annotation_count_ratio_df(task_history_df, task_df)
            productivity_df = self.table_obj.create_productivity_per_user_from_aw_time(
                df_task_history=task_history_df,
                df_labor=self._get_labor_df(),
                df_worktime_ratio=annotation_count_ratio_df,
            )
            self.productivity_df = productivity_df
        return self.productivity_df

    def _get_whole_productivity_df(self):
        if self.whole_productivity_df is None:
            task_df = self._get_task_df()
            labor_df = self._get_labor_df()
            self.whole_productivity_df = self.table_obj.create_whole_productivity_per_date(task_df, labor_df)
        return self.whole_productivity_df

    def write_histogram_for_task(self) -> None:
        """
        タスクに関するヒストグラムを出力する。

        """
        task_df = self._get_task_df()
        catch_exception(self.histogram_obj.write_histogram_for_worktime)(task_df)
        catch_exception(self.histogram_obj.write_histogram_for_other)(task_df)

        if not self.minimal_output:
            catch_exception(self.histogram_obj.write_histogram_for_annotation_worktime_by_user)(task_df)
            catch_exception(self.histogram_obj.write_histogram_for_inspection_worktime_by_user)(task_df)
            catch_exception(self.histogram_obj.write_histogram_for_acceptance_worktime_by_user)(task_df)

    def write_histogram_for_annotation(self) -> None:
        """
        アノテーションに関するヒストグラムを出力する。
        """
        annotation_df = self._get_annotation_df()
        catch_exception(self.histogram_obj.write_histogram_for_annotation_count_by_label)(annotation_df)

    def write_scatter_per_user(self) -> None:
        """
        ユーザごとにプロットした散布図を出力する。
        """
        productivity_df = self._get_productivity_df()
        catch_exception(self.scatter_obj.write_scatter_for_productivity_by_monitored_worktime)(productivity_df)
        catch_exception(self.scatter_obj.write_scatter_for_productivity_by_actual_worktime)(productivity_df)
        catch_exception(self.scatter_obj.write_scatter_for_quality)(productivity_df)
        catch_exception(self.scatter_obj.write_scatter_for_productivity_by_actual_worktime_and_quality)(productivity_df)

    def write_linegraph_for_task_overall(self) -> None:
        """
        タスク関係の折れ線グラフを出力する。

        Args:
            user_id_list: 折れ線グラフに表示するユーザ

        Returns:

        """
        task_df = self._get_task_df()
        task_cumulative_df_overall = Table.create_cumulative_df_overall(task_df)
        catch_exception(self.linegraph_obj.write_cumulative_line_graph_overall)(task_cumulative_df_overall)

    def write_whole_linegraph(self) -> None:
        whole_productivity_df = self._get_whole_productivity_df()
        catch_exception(self.linegraph_obj.write_whole_productivity_line_graph)(whole_productivity_df)
        catch_exception(self.linegraph_obj.write_whole_cumulative_line_graph)(whole_productivity_df)

    def write_linegraph_by_user(self, user_id_list: Optional[List[str]] = None) -> None:
        """
        折れ線グラフをユーザごとにプロットする。

        Args:
            user_id_list: 折れ線グラフに表示するユーザ

        Returns:

        """
        task_df = self._get_task_df()
        if len(task_df) == 0:
            logger.warning(f"タスク一覧が0件のため、折れ線グラフを出力しません。")
            return

        output_target_list: Optional[List[OutputTarget]] = None
        if self.minimal_output:
            output_target_list = [OutputTarget.ANNOTATION]

        # 単位あたりの指標を算出
        task_df = Table.create_gradient_df(task_df)

        task_cumulative_df_by_annotator = self.table_obj.create_cumulative_df_by_first_annotator(task_df)
        catch_exception(self.linegraph_obj.write_cumulative_line_graph_for_annotator)(
            df=task_cumulative_df_by_annotator,
            output_target_list=output_target_list,
            first_annotation_user_id_list=user_id_list,
        )

        task_cumulative_df_by_inspector = self.table_obj.create_cumulative_df_by_first_inspector(task_df)
        catch_exception(self.linegraph_obj.write_cumulative_line_graph_for_inspector)(
            df=task_cumulative_df_by_inspector,
            output_target_list=output_target_list,
            first_inspection_user_id_list=user_id_list,
        )

        task_cumulative_df_by_acceptor = self.table_obj.create_cumulative_df_by_first_acceptor(task_df)
        catch_exception(self.linegraph_obj.write_cumulative_line_graph_for_acceptor)(
            df=task_cumulative_df_by_acceptor,
            output_target_list=output_target_list,
            first_acceptance_user_id_list=user_id_list,
        )

        if not self.minimal_output:
            df_by_date_user_for_annotation = self._get_df_by_date_user_for_annotation()
            if len(df_by_date_user_for_annotation) > 0:
                catch_exception(self.linegraph_obj.write_productivity_line_graph_for_annotator)(
                    df=df_by_date_user_for_annotation, first_annotation_user_id_list=user_id_list
                )
                catch_exception(self.linegraph_obj.write_gradient_graph_for_annotator)(
                    df=task_cumulative_df_by_annotator, first_annotation_user_id_list=user_id_list
                )

            df_by_date_user_for_inspection = self._get_df_by_date_user_for_inspection()
            if len(df_by_date_user_for_inspection) > 0:
                catch_exception(self.linegraph_obj.write_productivity_line_graph_for_inspector)(
                    df=df_by_date_user_for_inspection, first_inspection_user_id_list=user_id_list
                )
                catch_exception(self.linegraph_obj.write_gradient_for_inspector)(
                    df=task_cumulative_df_by_inspector, first_inspection_user_id_list=user_id_list
                )

            df_by_date_user_for_acceptance = self._get_df_by_date_user_for_acceptance()
            if len(df_by_date_user_for_acceptance) > 0:
                catch_exception(self.linegraph_obj.write_productivity_line_graph_for_acceptor)(
                    df=df_by_date_user_for_acceptance, first_acceptance_user_id_list=user_id_list
                )
                catch_exception(self.linegraph_obj.write_gradient_for_acceptor)(
                    df=task_cumulative_df_by_acceptor, first_acceptance_user_id_list=user_id_list
                )

    def write_linegraph_for_worktime_by_user(self, user_id_list: Optional[List[str]] = None) -> None:
        account_statistics_df = self._get_account_statistics_df()
        cumulative_account_statistics_df = self.table_obj.create_cumulative_df_by_user(account_statistics_df)
        catch_exception(self.linegraph_obj.write_cumulative_line_graph_by_date)(
            df=cumulative_account_statistics_df, user_id_list=user_id_list
        )

    def write_csv_for_task(self) -> None:
        """
        タスク関係のCSVを出力する。
        """
        task_df = self._get_task_df()
        catch_exception(self.csv_obj.write_task_list)(task_df, dropped_columns=["input_data_id_list"])

    def write_csv_for_summary(self) -> None:
        """
        タスク関係のCSVを出力する。
        """
        task_df = self._get_task_df()
        catch_exception(self.csv_obj.write_task_count_summary)(task_df)
        catch_exception(self.csv_obj.write_worktime_summary)(task_df)
        catch_exception(self.csv_obj.write_count_summary)(task_df)

    def write_whole_productivity_csv_per_date(self) -> None:
        """
        日毎の生産性を出力する
        """
        whole_productivity_df = self._get_whole_productivity_df()
        catch_exception(self.csv_obj.write_whole_productivity_per_date)(whole_productivity_df)

    def _write_メンバー別作業時間平均_画像1枚あたり_by_phase(self, phase: TaskPhase):
        df_by_inputs = self.table_obj.create_worktime_per_image_df(AggregationBy.BY_INPUTS, phase)
        self.csv_obj.write_メンバー別作業時間平均_画像1枚あたり(df_by_inputs, phase)

        df_by_tasks = self.table_obj.create_worktime_per_image_df(AggregationBy.BY_TASKS, phase)
        self.csv_obj.write_メンバー別作業時間平均_タスク1個あたり(df_by_tasks, phase)

    def write_メンバー別作業時間平均_画像1枚あたり_by_phase(self):
        for phase in TaskPhase:
            catch_exception(self._write_メンバー別作業時間平均_画像1枚あたり_by_phase)(phase)

    def write_csv_for_inspection(self) -> None:
        """
        検査コメント関係の情報をCSVに出力する。
        """
        inspection_df = self.table_obj.create_inspection_df()
        inspection_df_all = self.table_obj.create_inspection_df(only_error_corrected=False)

        catch_exception(self.csv_obj.write_inspection_list)(
            df=inspection_df, dropped_columns=["data"], only_error_corrected=True
        )
        catch_exception(self.csv_obj.write_inspection_list)(
            df=inspection_df_all,
            dropped_columns=["data"],
            only_error_corrected=False,
        )

    def write_csv_for_annotation(self) -> None:
        """
        アノテーション関係の情報をCSVに出力する。
        """
        annotation_df = self._get_annotation_df()
        catch_exception(self.csv_obj.write_ラベルごとのアノテーション数)(annotation_df)

    def write_csv_for_account_statistics(self) -> None:
        account_statistics_df = self._get_account_statistics_df()
        catch_exception(self.csv_obj.write_ユーザ別日毎の作業時間)(account_statistics_df)

    def write_csv_for_date_user(self) -> None:
        """
        ユーザごと、日ごとの情報をCSVに出力する.
        """
        df_by_date_user = self._get_df_by_date_user_for_annotation()
        catch_exception(self.csv_obj.write_教師付作業者別日毎の情報)(df_by_date_user)

    def write_productivity_csv_per_user(self) -> None:
        productivity_df = self._get_productivity_df()
        catch_exception(self.csv_obj.write_productivity_per_user)(productivity_df)
        catch_exception(self.csv_obj.write_whole_productivity)(productivity_df)

    def write_labor_and_task_history(self) -> None:
        task_history_df = self._get_task_history_df()
        catch_exception(self.csv_obj.write_task_history_list)(task_history_df)

        df_labor = self._get_labor_df()
        catch_exception(self.csv_obj.write_labor_list)(df_labor)


def visualize_statistics(
    project_id: str,
    annofab_service: annofabapi.Resource,
    annofab_facade: AnnofabApiFacade,
    work_dir: Path,
    output_project_dir: Path,
    task_query: Dict[str, Any],
    ignored_task_id_list: Optional[List[str]],
    user_id_list: Optional[List[str]],
    update: bool = False,
    download_latest: bool = False,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    minimal_output: bool = False,
):
    """
    プロジェクトの統計情報を出力する。

    Args:
        project_id: 対象のproject_id
        task_query: タスク検索クエリ

    """

    annofab_facade.validate_project(
        project_id, project_member_roles=[ProjectMemberRole.OWNER, ProjectMemberRole.TRAINING_DATA_USER]
    )

    checkpoint_dir = work_dir / project_id
    checkpoint_dir.mkdir(exist_ok=True, parents=True)

    database = Database(
        annofab_service,
        project_id,
        str(checkpoint_dir),
        query=Query(
            task_query_param=task_query,
            ignored_task_id_list=ignored_task_id_list,
            start_date=start_date,
            end_date=end_date,
        ),
    )
    if update:
        database.update_db(download_latest)

    table_obj = Table(database, ignored_task_id_list)
    if len(table_obj._get_task_list()) == 0:
        logger.warning(f"project_id={project_id}: タスク一覧が0件なのでファイルを出力しません。終了します。")
        return
    if len(table_obj._get_task_histories_dict().keys()) == 0:
        logger.warning(f"project_id={project_id}: タスク履歴一覧が0件なのでファイルを出力しません。終了します。")
        return

    write_project_name_file(
        annofab_service,
        project_id=project_id,
        command_line_args=CommnadLineArgs(
            task_query=task_query,
            user_id_list=user_id_list,
            start_date=start_date,
            end_date=end_date,
            ignored_task_id_list=ignored_task_id_list,
        ),
        output_project_dir=output_project_dir,
    )
    write_obj = WriteCsvGraph(table_obj, output_project_dir, minimal_output)
    write_obj.write_csv_for_task()

    write_obj.write_csv_for_summary()
    write_obj.write_whole_productivity_csv_per_date()
    write_obj.write_productivity_csv_per_user()

    # 散布図
    write_obj.write_scatter_per_user()

    # ヒストグラム
    write_obj.write_histogram_for_task()

    # 折れ線グラフ
    write_obj.write_linegraph_by_user(user_id_list)
    write_obj.write_whole_linegraph()

    if not minimal_output:
        write_obj.write_histogram_for_annotation()
        write_obj.write_linegraph_for_worktime_by_user(user_id_list)

        # CSV
        write_obj.write_labor_and_task_history()
        write_obj.write_csv_for_annotation()
        write_obj.write_csv_for_account_statistics()
        write_obj.write_csv_for_date_user()
        write_obj.write_csv_for_inspection()
        write_obj.write_メンバー別作業時間平均_画像1枚あたり_by_phase()


def visualize_statistics_wrapper(
    project_id: str,
    root_output_dir: Path,
    annofab_service: annofabapi.Resource,
    annofab_facade: AnnofabApiFacade,
    work_dir: Path,
    task_query: Dict[str, Any],
    ignored_task_id_list: Optional[List[str]],
    user_id_list: Optional[List[str]],
    update: bool = False,
    download_latest: bool = False,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    minimal_output: bool = False,
) -> Optional[Path]:

    try:
        project_title = annofab_facade.get_project_title(project_id)
        output_project_dir = root_output_dir / get_project_output_dir(project_title)

        visualize_statistics(
            annofab_service=annofab_service,
            annofab_facade=annofab_facade,
            project_id=project_id,
            work_dir=work_dir,
            output_project_dir=output_project_dir,
            task_query=task_query,
            ignored_task_id_list=ignored_task_id_list,
            user_id_list=user_id_list,
            update=update,
            download_latest=download_latest,
            start_date=start_date,
            end_date=end_date,
            minimal_output=minimal_output,
        )
        return output_project_dir
    except Exception:  # pylint: disable=broad-except
        logger.warning(f"project_id={project_id}の統計情報の出力に失敗しました。", exc_info=True)
        return None


class VisualizeStatistics(AbstractCommandLineInterface):
    """
    統計情報を可視化する。
    """

    def visualize_statistics_for_project_list(
        self,
        root_output_dir: Path,
        project_id_list: List[str],
        work_dir: Path,
        task_query: Dict[str, Any],
        ignored_task_id_list: Optional[List[str]],
        user_id_list: Optional[List[str]],
        update: bool = False,
        download_latest: bool = False,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        minimal_output: bool = False,
        parallelism: Optional[int] = None,
    ) -> List[Path]:
        output_project_dir_list: List[Path] = []

        if parallelism is not None:
            partial_func = partial(
                visualize_statistics_wrapper,
                root_output_dir=root_output_dir,
                annofab_service=self.service,
                annofab_facade=self.facade,
                work_dir=work_dir,
                task_query=task_query,
                ignored_task_id_list=ignored_task_id_list,
                user_id_list=user_id_list,
                update=update,
                download_latest=download_latest,
                start_date=start_date,
                end_date=end_date,
                minimal_output=minimal_output,
            )
            with Pool(parallelism) as pool:
                result_list = pool.map(partial_func, project_id_list)
                output_project_dir_list = [e for e in result_list if e is not None]

        else:

            for project_id in project_id_list:
                output_project_dir = visualize_statistics_wrapper(
                    root_output_dir=root_output_dir,
                    annofab_service=self.service,
                    annofab_facade=self.facade,
                    project_id=project_id,
                    work_dir=work_dir,
                    task_query=task_query,
                    ignored_task_id_list=ignored_task_id_list,
                    user_id_list=user_id_list,
                    update=update,
                    download_latest=download_latest,
                    start_date=start_date,
                    end_date=end_date,
                    minimal_output=minimal_output,
                )
                if output_project_dir is not None:
                    output_project_dir_list.append(output_project_dir)

        return output_project_dir_list

    def main(self):
        args = self.args
        task_query = annofabcli.common.cli.get_json_from_args(args.task_query)
        ignored_task_id_list = (
            annofabcli.common.cli.get_list_from_args(args.ignored_task_id) if args.ignored_task_id is not None else None
        )
        user_id_list = annofabcli.common.cli.get_list_from_args(args.user_id) if args.user_id is not None else None
        project_id_list = annofabcli.common.cli.get_list_from_args(args.project_id)

        if args.work_dir is not None:
            work_dir = args.work_dir
        else:
            work_dir = annofabcli.utils.get_cache_dir()

        root_output_dir: Path = args.output_dir

        if len(project_id_list) == 1:
            visualize_statistics(
                annofab_service=self.service,
                annofab_facade=self.facade,
                project_id=project_id_list[0],
                output_project_dir=root_output_dir,
                work_dir=work_dir,
                task_query=task_query,
                ignored_task_id_list=ignored_task_id_list,
                user_id_list=user_id_list,
                update=not args.not_update,
                download_latest=args.download_latest,
                start_date=args.start_date,
                end_date=args.end_date,
                minimal_output=args.minimal,
            )

        else:
            # project_idが複数指定された場合は、project_titleのディレクトリに統計情報を出力する
            output_project_dir_list = self.visualize_statistics_for_project_list(
                root_output_dir=root_output_dir,
                project_id_list=project_id_list,
                work_dir=work_dir,
                task_query=task_query,
                ignored_task_id_list=ignored_task_id_list,
                user_id_list=user_id_list,
                update=not args.not_update,
                download_latest=args.download_latest,
                start_date=args.start_date,
                end_date=args.end_date,
                minimal_output=args.minimal,
                parallelism=args.parallelism,
            )

            if args.merge:
                if len(output_project_dir_list) > 1:
                    merge_visualization_dir(
                        project_dir_list=output_project_dir_list,
                        output_dir=root_output_dir / "merge",
                        user_id_list=user_id_list,
                        minimal_output=args.minimal,
                    )
                else:
                    logger.warning(f"出力した統計情報は1件以下なので、`merge`ディレクトリを出力しません。")

            if len(output_project_dir_list) > 0:
                whole_peformance_csv_list = [e / FILENAME_WHOLE_PEFORMANCE for e in output_project_dir_list]
                write_summarise_whole_peformance_csv(
                    csv_path_list=whole_peformance_csv_list, output_path=root_output_dir / "プロジェクトごとの生産性と品質.csv"
                )
            else:
                logger.warning(f"出力した統計情報は0なので、`プロジェクトごとの生産性と品質.csv`を出力しません。")


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    VisualizeStatistics(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):

    parser.add_argument(
        "-p",
        "--project_id",
        required=True,
        nargs="+",
        help=(
            "対象のプロジェクトのproject_idを指定してください。複数指定した場合、プロジェクトごとに統計情報が出力されます。"
            "file://`を先頭に付けると、project_idが記載されたファイルを指定できます。"
        ),
    )

    parser.add_argument("-o", "--output_dir", type=Path, required=True, help="出力先ディレクトリのパスを指定してください。")

    parser.add_argument(
        "-u",
        "--user_id",
        nargs="+",
        help=("メンバごとの統計グラフに表示するユーザのuser_idを指定してください。" "指定しない場合は、上位20人が表示されます。" "file://`を先頭に付けると、一覧が記載されたファイルを指定できます。"),
    )

    parser.add_argument(
        "-tq",
        "--task_query",
        type=str,
        help="タスクの検索クエリをJSON形式で指定します。指定しない場合はすべてのタスクを取得します。"
        "`file://`を先頭に付けると、JSON形式のファイルを指定できます。"
        "クエリのキーは、phase, status, task_id のみです。[getTasks API](https://annofab.com/docs/api/#operation/getTasks) 参照",
    )

    parser.add_argument("--start_date", type=str, help="指定した日付（'YYYY-MM-DD'）以降に教師付を開始したタスクを集計する。")
    parser.add_argument("--end_date", type=str, help="指定した日付（'YYYY-MM-DD'）以前に更新されたタスクを集計する。")

    parser.add_argument(
        "--ignored_task_id",
        nargs="+",
        help=("可視化対象外のタスクのtask_id。" "指定しない場合は、すべてのタスクが可視化対象です。" "file://`を先頭に付けると、一覧が記載されたファイルを指定できます。"),
    )

    parser.add_argument(
        "--not_update",
        action="store_true",
        help="作業ディレクトリ内のファイルを参照して、統計情報を出力します。" "AnnoFab Web APIへのアクセスを最小限にします。",
    )

    parser.add_argument(
        "--download_latest",
        action="store_true",
        help="統計情報の元になるファイル（アノテーションzipなど）の最新版をダウンロードします。ファイルを最新版にするのに5分以上待つ必要があります。",
    )

    parser.add_argument(
        "--work_dir",
        type=Path,
        help="作業ディレクトリのパス。指定しない場合はannofabcliのキャッシュディレクトリ（'$HOME/.cache/annofabcli'）に保存します。",
    )

    parser.add_argument(
        "--minimal",
        action="store_true",
        help="必要最小限のファイルを出力します。",
    )

    parser.add_argument(
        "--merge",
        action="store_true",
        help="指定した場合、複数のproject_idを指定したときに、マージした統計情報も出力します。ディレクトリ名は`merge`です。",
    )

    parser.add_argument(
        "--parallelism",
        type=int,
        help="並列度。`--project_id`に複数のproject_idを指定したときのみ有効なオプションです。指定しない場合は、逐次的に処理します。また、必ず'--yes'を指定してください。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "visualize"
    subcommand_help = "統計情報を可視化したファイルを出力します。"
    description = "統計情報を可視化したファイルを出力します。毎日 03:00JST頃に更新されます。"
    epilog = "アノテーションユーザまたはオーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import logging
import multiprocessing
import sys
from functools import partial
from typing import List, Optional, Tuple

import annofabapi
import requests
from annofabapi.dataclass.task import Task
from annofabapi.models import ProjectMemberRole, TaskStatus

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    prompt_yesnoall,
)
from annofabcli.common.facade import TaskQuery, match_task_with_query

logger = logging.getLogger(__name__)


class ChangeOperatorMain:
    def __init__(self, service: annofabapi.Resource, all_yes: bool):
        self.service = service
        self.facade = AnnofabApiFacade(service)
        self.all_yes = all_yes

    def confirm_processing(self, confirm_message: str) -> bool:
        """
        `all_yes`属性を見て、処理するかどうかユーザに問い合わせる。
        "ALL"が入力されたら、`all_yes`属性をTrueにする

        Args:
            task_id: 処理するtask_id
            confirm_message: 確認メッセージ

        Returns:
            True: Yes, False: No

        """
        if self.all_yes:
            return True

        yes, all_yes = prompt_yesnoall(confirm_message)

        if all_yes:
            self.all_yes = True

        return yes

    def confirm_change_operator(self, task: Task) -> bool:
        confirm_message = f"task_id = {task.task_id} のタスクの担当者を変更しますか？"
        return self.confirm_processing(confirm_message)

    def change_operator_for_task(
        self,
        project_id: str,
        task_id: str,
        new_account_id: Optional[str] = None,
        task_query: Optional[TaskQuery] = None,
        task_index: Optional[int] = None,
    ) -> bool:

        logging_prefix = f"{task_index+1} 件目" if task_index is not None else ""
        dict_task, _ = self.service.api.get_task(project_id, task_id)
        task: Task = Task.from_dict(dict_task)

        now_user_id = None
        if task.account_id is not None:
            now_user_id = self.facade.get_user_id_from_account_id(project_id, task.account_id)

        logger.debug(
            f"{logging_prefix} : task_id = {task.task_id}, "
            f"status = {task.status.value}, "
            f"phase = {task.phase.value}, "
            f"phase_stage = {task.phase_stage}, "
            f"user_id = {now_user_id}"
        )

        if task.status in [TaskStatus.COMPLETE, TaskStatus.WORKING]:
            logger.warning(f"{logging_prefix} : task_id = {task_id} : タスクのstatusがworking or complete なので、担当者を変更できません。")
            return False

        if not match_task_with_query(task, task_query):
            logger.debug(f"{logging_prefix} : task_id = {task_id} : TaskQueryの条件にマッチしないため、スキップします。")
            return False

        if not self.confirm_change_operator(task):
            return False

        try:
            # 担当者を変更する
            self.facade.change_operator_of_task(project_id, task_id, new_account_id)
            logger.debug(f"{logging_prefix} : task_id = {task_id}, phase={dict_task['phase']} のタスクの担当者を変更しました。")
            return True

        except requests.exceptions.HTTPError as e:
            logger.warning(f"{logging_prefix} : task_id = {task_id} の担当者を変更するのに失敗しました。")
            logger.warning(e)
            return False

    def change_operator_for_task_wrapper(
        self,
        tpl: Tuple[int, str],
        project_id: str,
        task_query: Optional[TaskQuery] = None,
        new_account_id: Optional[str] = None,
    ) -> bool:
        task_index, task_id = tpl
        return self.change_operator_for_task(
            project_id=project_id,
            task_id=task_id,
            task_index=task_index,
            task_query=task_query,
            new_account_id=new_account_id,
        )

    def change_operator(
        self,
        project_id: str,
        task_id_list: List[str],
        new_user_id: Optional[str] = None,
        task_query: Optional[TaskQuery] = None,
        parallelism: Optional[int] = None,
    ):
        """
        検査コメントを付与して、タスクを差し戻す
        Args:
            project_id:
            task_id_list:
            new_user_id: 新しく担当するユーザのuser_id。Noneの場合タスクの担当者は未割り当てにする。

        """
        if task_query is not None:
            task_query = self.facade.set_account_id_of_task_query(project_id, task_query)

        if new_user_id is not None:
            new_account_id = self.facade.get_account_id_from_user_id(project_id, new_user_id)
            if new_account_id is None:
                logger.error(f"ユーザ '{new_user_id}' のaccount_idが見つかりませんでした。終了します。")
                return
            else:
                logger.info(f" {len(task_id_list)}のタスクの担当者を、{new_user_id}に変更します。")
        else:
            new_account_id = None
            logger.info(f" {len(task_id_list)} 件のタスクの担当者を未割り当てに変更します。")

        success_count = 0

        if parallelism is not None:
            partial_func = partial(
                self.change_operator_for_task_wrapper,
                project_id=project_id,
                task_query=task_query,
                new_account_id=new_account_id,
            )
            with multiprocessing.Pool(parallelism) as pool:
                result_bool_list = pool.map(partial_func, enumerate(task_id_list))
                success_count = len([e for e in result_bool_list if e])

        else:
            # 逐次処理
            for task_index, task_id in enumerate(task_id_list):
                result = self.change_operator_for_task(
                    project_id, task_id, task_index=task_index, task_query=task_query, new_account_id=new_account_id
                )
                if result:
                    success_count += 1

        logger.info(f"{success_count} / {len(task_id_list)} 件 タスクの担当者を変更しました。")


class ChangeOperator(AbstractCommandLineInterface):
    @staticmethod
    def validate(args: argparse.Namespace) -> bool:
        COMMON_MESSAGE = "annofabcli task change_operator: error:"

        if args.parallelism is not None and not args.yes:
            print(
                f"{COMMON_MESSAGE} argument --parallelism: '--parallelism'を指定するときは、必ず'--yes'を指定してください。",
                file=sys.stderr,
            )
            return False

        return True

    def main(self):
        args = self.args
        if not self.validate(args):
            return

        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id)
        if args.user_id is not None:
            user_id = args.user_id
        elif args.not_assign:
            user_id = None
        else:
            logger.error(f"タスクの担当者の指定方法が正しくありません。")
            return

        dict_task_query = annofabcli.common.cli.get_json_from_args(args.task_query)
        task_query: Optional[TaskQuery] = TaskQuery.from_dict(dict_task_query) if dict_task_query is not None else None

        project_id = args.project_id
        super().validate_project(project_id, [ProjectMemberRole.OWNER, ProjectMemberRole.ACCEPTER])

        main_obj = ChangeOperatorMain(self.service, all_yes=self.all_yes)
        main_obj.change_operator(
            project_id,
            task_id_list=task_id_list,
            new_user_id=user_id,
            task_query=task_query,
            parallelism=args.parallelism,
        )


def main(args: argparse.Namespace):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ChangeOperator(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    argument_parser.add_task_id()

    assign_group = parser.add_mutually_exclusive_group(required=True)

    assign_group.add_argument("-u", "--user_id", type=str, help="タスクを新しく担当するユーザのuser_idを指定してください。")

    assign_group.add_argument("--not_assign", action="store_true", help="指定した場合、タスクの担当者は未割り当てになります。")

    argument_parser.add_task_query()

    parser.add_argument(
        "--parallelism", type=int, help="使用するプロセス数（並列度）を指定してください。指定する場合は必ず'--yes'を指定してください。指定しない場合は、逐次的に処理します。"
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "change_operator"
    subcommand_help = "タスクの担当者を変更します。"
    description = "タスクの担当者を変更します。ただし、作業中/受入完了状態のタスクに対しては変更できません。"
    epilog = "チェッカーまたはオーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import logging
from pathlib import Path
from typing import Dict, List, Optional, Set

import pandas

import annofabcli
from annofabcli.common.cli import get_list_from_args
from annofabcli.common.utils import print_csv, read_multiheader_csv
from annofabcli.experimental.mask_user_info import (
    create_masked_user_info_df,
    create_replacement_dict_by_user_id,
    replace_by_columns,
)
from annofabcli.experimental.write_linegraph_per_user import write_linegraph_per_user
from annofabcli.experimental.write_performance_scatter_per_user import write_performance_scatter_per_user
from annofabcli.statistics.csv import FILENAME_PEFORMANCE_PER_USER, FILENAME_TASK_LIST

logger = logging.getLogger(__name__)


def _replace_df_task(df, replacement_dict_by_user_id: Dict[str, str]):
    replace_by_columns(df, replacement_dict_by_user_id, main_column="user_id", sub_columns=["username"])

    replace_by_columns(
        df,
        replacement_dict_by_user_id,
        main_column="first_annotation_user_id",
        sub_columns=["first_annotation_username"],
    )
    replace_by_columns(
        df,
        replacement_dict_by_user_id,
        main_column="first_inspection_user_id",
        sub_columns=["first_inspection_username"],
    )
    replace_by_columns(
        df,
        replacement_dict_by_user_id,
        main_column="first_acceptance_user_id",
        sub_columns=["first_acceptance_username"],
    )


def mask_visualization_dir(
    project_dir: Path,
    output_dir: Path,
    not_masked_biography_set: Optional[Set[str]] = None,
    not_masked_user_id_set: Optional[Set[str]] = None,
    minimal_output: bool = False,
    exclude_masked_user_for_linegraph: bool = False,
):
    df_member_perfomance = read_multiheader_csv(str(project_dir / FILENAME_PEFORMANCE_PER_USER), header_row_count=2)

    replacement_dict_by_user_id = create_replacement_dict_by_user_id(
        df_member_perfomance,
        not_masked_biography_set=not_masked_biography_set,
        not_masked_user_id_set=not_masked_user_id_set,
    )
    not_masked_user_id_set = set(df_member_perfomance[("user_id", "")]) - set(replacement_dict_by_user_id.keys())

    # CSVのユーザ情報をマスクする
    masked_df_member_perfomance = create_masked_user_info_df(
        df_member_perfomance,
        not_masked_biography_set=not_masked_biography_set,
        not_masked_user_id_set=not_masked_user_id_set,
    )
    print_csv(masked_df_member_perfomance, output=str(output_dir / FILENAME_PEFORMANCE_PER_USER))

    df_task = pandas.read_csv(str(project_dir / FILENAME_TASK_LIST))
    _replace_df_task(df_task, replacement_dict_by_user_id=replacement_dict_by_user_id)
    print_csv(df_task, output=str(output_dir / FILENAME_TASK_LIST))

    # メンバのパフォーマンスを散布図で出力する
    write_performance_scatter_per_user(output_dir / FILENAME_PEFORMANCE_PER_USER, output_dir=output_dir / "scatter")

    user_id_list: Optional[List[str]] = None
    if exclude_masked_user_for_linegraph:
        user_id_list = list(not_masked_user_id_set)

    # メンバごとにパフォーマンスを折れ線グラフで出力する
    write_linegraph_per_user(
        output_dir / FILENAME_TASK_LIST,
        output_dir=output_dir / "line-graph",
        minimal_output=minimal_output,
        user_id_list=user_id_list,
    )


def main(args):
    not_masked_biography_set = (
        set(get_list_from_args(args.not_masked_biography)) if args.not_masked_biography is not None else None
    )
    not_masked_user_id_set = (
        set(get_list_from_args(args.not_masked_user_id)) if args.not_masked_user_id is not None else None
    )

    mask_visualization_dir(
        project_dir=args.dir,
        not_masked_biography_set=not_masked_biography_set,
        not_masked_user_id_set=not_masked_user_id_set,
        output_dir=args.output_dir,
        minimal_output=args.minimal,
        exclude_masked_user_for_linegraph=args.exclude_masked_user_for_linegraph,
    )


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--dir", type=Path, required=True, help="マスク対象のディレクトリ。`annofabcli statistics visualize`コマンドの出力結果。"
    )

    parser.add_argument(
        "--not_masked_biography",
        type=str,
        nargs="+",
        help="マスクしないユーザの`biography`を指定してください。",
    )

    parser.add_argument(
        "--not_masked_user_id",
        type=str,
        nargs="+",
        help="マスクしないユーザの`user_id`を指定してください。",
    )

    parser.add_argument(
        "--minimal",
        action="store_true",
        help="必要最小限のファイルを出力します。",
    )

    parser.add_argument(
        "--exclude_masked_user_for_linegraph",
        action="store_true",
        help="折れ線グラフに、マスクされたユーザを除外します。",
    )

    parser.add_argument("-o", "--output_dir", type=Path, required=True, help="出力先ディレクトリ。")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "mask_visualization_dir"
    subcommand_help = "`annofabcli statistics visualize`コマンドの出力結果のユーザ情報をマスクします。"
    description = "`annofabcli statistics visualize`コマンドの出力結果のユーザ情報をマスクします。マスク対象のファイルのみ出力します。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)
